"""
importmap_dialog.py

This is the pop-up dialog for the custom map import function.

Created on 14 Sep 2024

:author: semuadmin (Steve Smith)
:copyright: 2020 semuadmin
:license: BSD 3-Clause
"""

# pylint: disable=unused-argument

from tkinter import (
    ALL,
    DISABLED,
    NORMAL,
    Button,
    Checkbutton,
    E,
    Entry,
    Frame,
    IntVar,
    Label,
    N,
    S,
    StringVar,
    W,
)

try:
    from rasterio import open as openraster  # pylint: disable=import-error
    from rasterio.warp import transform_bounds  # pylint: disable=import-error

    HASRASTERIO = True
except (ImportError, ModuleNotFoundError):
    HASRASTERIO = False

from pygpsclient.globals import (
    BGCOL,
    ERRCOL,
    IMPORT,
    INFOCOL,
    Area,
)
from pygpsclient.map_canvas import MapCanvas
from pygpsclient.strings import DLGTIMPORTMAP
from pygpsclient.toplevel_dialog import ToplevelDialog

# profile chart parameters:
AXIS_XL = 35  # x axis left offset
AXIS_XR = 35  # x axis right offset
AXIS_Y = 15  # y axis bottom offset
ELEAX_COL = "green4"  # color of elevation plot axis
ELE_COL = "palegreen3"  # color of elevation plot
SPD_COL = INFOCOL  # color of speed plot
MD_LINES = 2  # number of lines of metadata
MINDIM = (456, 418)


class ImportMapDialog(ToplevelDialog):
    """ImportMapDialog class."""

    def __init__(self, app, *args, **kwargs):
        """Constructor."""

        self.__app = app
        # self.__master = self.__app.appmaster  # link to root Tk window
        super().__init__(app, DLGTIMPORTMAP, MINDIM)
        self.width = int(kwargs.get("width", 400))
        self.height = int(kwargs.get("height", 400))
        self._first = IntVar()
        self._lonmin = StringVar()
        self._latmin = StringVar()
        self._lonmax = StringVar()
        self._latmax = StringVar()
        self._custommap = ""

        self._body()
        self._do_layout()
        self._attach_events()
        self._reset()
        self._finalise()

    def _body(self):
        """
        Create widgets.
        """

        self._frm_body = Frame(self.container, borderwidth=2, relief="groove")
        self._can_mapview = MapCanvas(
            self.__app, self._frm_body, width=self.width, height=self.height, bg=BGCOL
        )
        self._frm_controls = Frame(self._frm_body, borderwidth=2, relief="groove")
        self._btn_load = Button(
            self._frm_controls,
            image=self.img_load,
            width=40,
            command=self._on_load,
        )
        self._btn_import = Button(
            self._frm_controls,
            image=self.img_send,
            width=40,
            command=self._on_import,
        )
        self._btn_redraw = Button(
            self._frm_controls,
            image=self.img_redraw,
            width=40,
            command=self._on_redraw,
        )
        self._chk_first = Checkbutton(
            self._frm_controls, text="First?", variable=self._first
        )
        self._lbl_min = Label(self._frm_controls, text="Minimum:")
        self._lbl_minlat = Label(self._frm_controls, text="Latitude")
        self._ent_minlat = Entry(
            self._frm_controls, width=10, textvariable=self._latmin
        )
        self._lbl_minlon = Label(self._frm_controls, text="Longitude")
        self._ent_minlon = Entry(
            self._frm_controls, width=10, textvariable=self._lonmin
        )
        self._lbl_max = Label(self._frm_controls, text="Maximum:")
        self._lbl_maxlat = Label(self._frm_controls, text="Latitude")
        self._ent_maxlat = Entry(
            self._frm_controls, width=10, textvariable=self._latmax
        )
        self._lbl_maxlon = Label(self._frm_controls, text="Longitude")
        self._ent_maxlon = Entry(
            self._frm_controls, width=10, textvariable=self._lonmax
        )

    def _do_layout(self):
        """
        Arrange widgets.
        """
        self._frm_body.grid(column=0, row=0, sticky=(N, S, E, W))
        self._can_mapview.grid(column=0, row=0, sticky=(N, S, E, W))
        self._frm_controls.grid(column=0, row=1, sticky=(W, E))
        self._btn_load.grid(column=0, row=0, padx=3, pady=3)
        self._btn_redraw.grid(column=1, row=0, padx=3, pady=3)
        self._btn_import.grid(column=2, row=0, padx=3, pady=3)
        self._chk_first.grid(column=3, row=0, padx=3, pady=3)
        self._lbl_min.grid(column=0, row=1, padx=3, pady=3, sticky=W)
        self._lbl_minlat.grid(column=1, row=1, padx=3, pady=3, sticky=E)
        self._ent_minlat.grid(column=2, row=1, padx=3, pady=3)
        self._lbl_minlon.grid(column=3, row=1, padx=3, pady=3, sticky=E)
        self._ent_minlon.grid(column=4, row=1, padx=3, pady=3)
        self._lbl_max.grid(column=0, row=2, padx=3, pady=3, sticky=W)
        self._lbl_maxlat.grid(column=1, row=2, padx=3, pady=3, sticky=E)
        self._ent_maxlat.grid(column=2, row=2, padx=3, pady=3)
        self._lbl_maxlon.grid(column=3, row=2, padx=3, pady=3, sticky=E)
        self._ent_maxlon.grid(column=4, row=2, padx=3, pady=3)
        self.container.grid_columnconfigure(0, weight=10)
        self.container.grid_rowconfigure(0, weight=10)
        self.grid_columnconfigure(0, weight=10)
        self.grid_rowconfigure(0, weight=10)
        self._frm_body.grid_columnconfigure(0, weight=10)
        self._frm_body.grid_rowconfigure(0, weight=10)
        self._can_mapview.grid_columnconfigure(0, weight=10)
        self._can_mapview.grid_rowconfigure(0, weight=10)

    def _attach_events(self):
        """
        Bind events to window.
        """

        self.container.bind("<Configure>", self._on_resize)

    def _reset(self):
        """
        Reset application.
        """

        self._first.set(0)
        self._can_mapview.delete(ALL)
        # self._btn_import.config(state=DISABLED)
        if not HASRASTERIO:
            self.set_status(
                "Warning: rasterio library is not installed - bounds must be entered manually",
                INFOCOL,
            )
        else:
            self.set_status("")

    def _on_load(self):
        """
        Load custom map from file.
        """

        self.set_status("")
        self._custommap = self._open_mapfile()
        if self._custommap is not None:
            bounds = self._get_bounds(self._custommap)
            self._can_mapview.draw_map(
                maptype=IMPORT, mappath=self._custommap, bounds=bounds, zoom=None
            )
            self._btn_import.config(state=NORMAL)
        else:
            self._btn_import.config(state=DISABLED)

    def _on_redraw(self, *args, **kwargs):
        """
        Handle redraw button press.
        """

        self._reset()
        self._can_mapview.draw_map(
            maptype=IMPORT,
            mappath=self._custommap,
            bounds=self._can_mapview.bounds,
            marker=self._can_mapview.marker,
            zoom=None,
        )

    def _open_mapfile(self) -> str:
        """
        Open custom map file.
        """

        return self.__app.file_handler.open_file(
            "tif",
            (("GeoTiff files", "*.tif"), ("all files", "*.*")),
        )

    def _get_bounds(self, mappath) -> Area:
        """
        Get bounds of custom map image file.

        :param str mappath: fully qualified path to map file
        :return: bounds (extent)
        :rtype: Area
        """

        lonmin = latmin = lonmax = latmax = 0.0
        if HASRASTERIO:
            try:
                ras = openraster(mappath)
                lonmin, latmin, lonmax, latmax = transform_bounds(
                    ras.crs.to_epsg(), 4326, *ras.bounds
                )
            except Exception:  # pylint: disable=broad-exception-caught
                self.set_status(
                    "Warning: image is not georeferenced - bounds must be entered manually",
                    ERRCOL,
                )

        self._lonmin.set(round(lonmin, 8))
        self._latmin.set(round(latmin, 8))
        self._lonmax.set(round(lonmax, 8))
        self._latmax.set(round(latmax, 8))
        return Area(
            float(self._latmin.get()),
            float(self._lonmin.get()),
            float(self._latmax.get()),
            float(self._lonmax.get()),
        )

    def _on_import(self):
        """
        Validate bounds and import custom file into saved config.
        """

        try:
            lonmin = float(self._lonmin.get())
            lonmax = float(self._lonmax.get())
            latmin = float(self._latmin.get())
            latmax = float(self._latmax.get())
        except ValueError:
            self.set_status("Error: invalid bounds", ERRCOL)
            return

        if lonmax + 180 <= lonmin + 180 or latmax + 90 <= latmin + 90:
            self.set_status("Error: minimum must be less than maximum", ERRCOL)
        else:
            usermaps = self.__app.configuration.get("usermaps_l")
            idx = 0 if self._first.get() else len(usermaps) + 1
            usermaps.insert(idx, [self._custommap, [latmin, lonmin, latmax, lonmax]])
            self.__app.configuration.set("usermaps_l", usermaps)
            self.set_status("Custom map imported", INFOCOL)



"""
scatter_frame.py

Scatterplot frame class for PyGPSClient Application.

This generates a scatterplot of positions, centered on either
the cumulative average position or a fixed reference position.

The fixed reference position can be stored in the json
configuration file as `scatterlat_f`/`scatterlon_f`.

Created 23 March 2023

Completely rewritten by semuadmin 5 Nov 2024 to use bounding
box rather than polar coordinations, and allow right-click
fixed reference selection.

:author: Nathan Michaels, semuadmin
:copyright: 2024 SEMU Consulting
:license: BSD 3-Clause
"""

from tkinter import (
    ALL,
    HORIZONTAL,
    NW,
    Canvas,
    Checkbutton,
    E,
    Entry,
    Frame,
    IntVar,
    N,
    S,
    Scale,
    Spinbox,
    StringVar,
    TclError,
    W,
)

try:
    from statistics import fmean, stdev

    HASSTATS = True
except (ImportError, ModuleNotFoundError):
    HASSTATS = False

from random import randrange

from pygpsclient.globals import (
    BGCOL,
    FGCOL,
    PNTCOL,
    READONLY,
    SQRT2,
    WIDGETU1,
    Area,
    Point,
)
from pygpsclient.helpers import (
    fontheight,
    get_point_at_vector,
    ll2xy,
    point_in_bounds,
    reorder_range,
    scale_font,
    xy2ll,
)

AVG = "avg"
CTRAVG = "Average"
CTRFIX = "Fixed"
CRTS = (CTRAVG, CTRFIX)
INTS = (1, 2, 5, 10, 20, 50, 100)
FIXCOL = "green2"
PNTTOPCOL = "red"
CULLMID = True  # whether to cull random points from middle of array
FIXINAUTO = False  # whether to include fixed ref point in autorange
MAXPOINTS = 2000
PNT = "pnt"
STDINT = 10  # standard deviation calculation interval


class ScatterViewFrame(Frame):
    """
    Scatterplot view frame class.
    """

    def __init__(self, app, *args, **kwargs):
        """
        Constructor.

        :param Frame app: reference to main tkinter application
        :param args: Optional args to pass to Frame parent class
        :param kwargs: Optional kwargs to pass to Frame parent class
        """
        self.__app = app
        self.__master = self.__app.appmaster

        Frame.__init__(self, self.__master, *args, **kwargs)

        def_w, def_h = WIDGETU1

        self.width = kwargs.get("width", def_w)
        self.height = kwargs.get("height", def_h)
        self._font = self.__app.font_sm
        self._fonth = fontheight(self._font)
        self._maxpoints = 0
        self._points = []
        self._average = None
        self._stddev = None
        self._fixed = None
        self._bounds = None
        self._minlat = 100
        self._minlon = 200
        self._maxlat = -100
        self._maxlon = -200
        self._updcount = -1
        self._lastbounds = Area(0, 0, 0, 0)
        self._range = 0.0
        self._autorange = IntVar()
        self._interval = IntVar()
        self._centermode = StringVar()
        self._scale = IntVar()
        self._reflat = StringVar()
        self._reflon = StringVar()
        self._scale_factors = (
            5000,
            2000,
            1000,
            500,
            200,
            100,
            50,
            20,
            10,
            5,
            2,
            1,
            0.5,
            0.2,
            0.1,
            0.05,
            0.02,
            0.01,
        )  # scale factors represent plot radius in meters

        self._body()
        self.reset()
        self._attach_events()

    def _body(self):
        """Set up frame and widgets."""

        for i in range(3):
            self.grid_columnconfigure(i, weight=1, uniform="ent")
        self.grid_rowconfigure(0, weight=1)
        self.grid_rowconfigure(1, weight=0)
        self.grid_rowconfigure(2, weight=0)
        self.canvas = Canvas(self, width=self.width, height=self.height, bg=BGCOL)
        self._ent_reflat = Entry(
            self,
            width=12,
            textvariable=self._reflat,
            fg=FIXCOL,
            bg=BGCOL,
        )
        self._ent_reflon = Entry(
            self,
            width=12,
            textvariable=self._reflon,
            fg=FIXCOL,
            bg=BGCOL,
        )
        self._chk_autorange = Checkbutton(
            self,
            text="Autorange",
            fg=PNTCOL,
            bg=BGCOL,
            variable=self._autorange,
        )
        crng = reorder_range(CRTS, self._centermode.get())
        self._spn_center = Spinbox(
            self,
            values=crng,
            width=8,
            wrap=True,
            background=BGCOL,
            readonlybackground=BGCOL,
            foreground=PNTCOL,
            buttonbackground=BGCOL,
            textvariable=self._centermode,
            state=READONLY,
        )
        irng = reorder_range(INTS, self._interval.get())
        self._spn_interval = Spinbox(
            self,
            values=irng,
            width=5,
            wrap=True,
            fg=PNTCOL,
            bg=BGCOL,
            readonlybackground=BGCOL,
            buttonbackground=BGCOL,
            textvariable=self._interval,
            state=READONLY,
        )
        self._scl_range = Scale(
            self,
            from_=0,
            to=len(self._scale_factors) - 1,
            orient=HORIZONTAL,
            bg=FGCOL,
            troughcolor=BGCOL,
            variable=self._scale,
            showvalue=False,
        )
        self.canvas.grid(column=0, row=0, columnspan=3, sticky=(N, S, E, W))
        self._ent_reflat.grid(column=0, row=1, sticky=(W, E))
        self._ent_reflon.grid(column=1, row=1, sticky=(W, E))
        self._spn_center.grid(column=2, row=1, sticky=(W, E))
        self._chk_autorange.grid(column=0, row=2, sticky=(W, E))
        self._spn_interval.grid(column=1, row=2, sticky=(W, E))
        self._scl_range.grid(column=2, row=2, sticky=(W, E))

    def reset(self):
        """
        Reset settings to saved configuration.
        """

        self._maxpoints = MAXPOINTS
        cfg = self.__app.configuration.get("scattersettings_d")
        reflat = cfg.get("scatterlat_f")
        reflon = cfg.get("scatterlon_f")
        self._reflat.set("Reference Lat" if reflat == 0.0 else reflat)
        self._reflon.set("Reference Lon" if reflon == 0.0 else reflon)
        self._autorange.set(cfg.get("scatterautorange_b"))
        self._scale.set(cfg.get("scatterscale_n"))
        self._interval.set(cfg.get("scatterinterval_n"))
        self._centermode.set(cfg.get("scattercenter_s"))
        if self._centermode.get() != CTRFIX:
            self._centermode.set(CTRAVG)

    def _attach_events(self):
        """
        Bind events to frame.
        """

        self.bind("<Configure>", self._on_resize)
        self.canvas.bind("<Double-Button-1>", self._on_clear)
        self.canvas.bind("<Button-2>", self._on_recenter)  # mac
        self.canvas.bind("<Button-3>", self._on_recenter)  # win
        self.canvas.bind("<MouseWheel>", self._on_zoom)
        self._scale.trace_add("write", self._on_rescale)
        for setting in (
            self._autorange,
            self._interval,
            self._centermode,
            self._reflat,
            self._reflon,
        ):
            setting.trace_add("write", self._on_update_config)

    def _on_zoom(self, event):
        """
        Adjust scale using mousewheel.

        :param event: mousewheel event
        """

        sl = len(self._scale_factors) - 1
        sc = self._scale.get()
        if event.delta > 0:
            if sc < sl:
                self._scale.set(sc + 1)
        elif event.delta < 0:
            if sc > 0:
                self._scale.set(sc - 1)

    def _on_rescale(self, var, index, mode):  # pylint: disable=unused-argument
        """
        Rescale widget.
        """

        self._on_update_config(var, index, mode)
        self._on_resize(None)

    def _on_recenter(self, event):
        """
        Right click centers on cursor.

        :param Event event: right click event
        """

        w, h = self.get_size()
        pos = xy2ll(w, h, self._bounds, (event.x, event.y))
        self._reflat.set(round(pos.lat, 9))
        self._reflon.set(round(pos.lon, 9))
        try:
            self._fixed = Point(float(self._reflat.get()), float(self._reflon.get()))
        except ValueError:
            pass

    def _on_clear(self, event):  # pylint: disable=unused-argument
        """ "
        Clear plot.

        :param Event event: double-click event
        """

        self._points = []
        self._minlat = 100
        self._minlon = 200
        self._maxlat = -100
        self._maxlon = -200
        self._updcount = -1
        self._init_frame()

    def _on_update_config(self, var, index, mode):  # pylint: disable=unused-argument
        """
        Update in-memory configuration if setting is changed.
        """

        try:
            self.update()
            sst = {}
            sst["maxpoints_n"] = int(self._maxpoints)
            sst["scatterautorange_b"] = int(self._autorange.get())
            sst["scattercenter_s"] = self._centermode.get()
            sst["scatterinterval_n"] = int(self._interval.get())
            sst["scatterscale_n"] = int(self._scale.get())
            sst["scatterlat_f"] = float(self._reflat.get())
            sst["scatterlon_f"] = float(self._reflon.get())
            self.__app.configuration.set("scattersettings_d", sst)
        except (ValueError, TclError):
            pass

    def _init_frame(self):
        """
        Initialize scatter plot.
        """

        width, height = self.get_size()
        self.canvas.delete(ALL)
        self.canvas.create_line(0, height / 2, width, height / 2, fill=FGCOL)
        self.canvas.create_line(width / 2, 0, width / 2, height, fill=FGCOL)

        if self._bounds is None:
            return

        maxr = min(height / 2, width / 2)
        for i in range(1, 5):
            self.canvas.create_circle(
                width / 2, height / 2, maxr * i / 4, outline=FGCOL, width=1
            )
            rng, unt = self.get_range_label()
            if rng >= 100:
                dp = 0
            elif rng >= 10:
                dp = 1
            elif rng >= 1:
                dp = 2
            else:
                dp = 3
            dist = f"{rng * i/4:.{dp}f}{unt}"
            txt_x = width / 2 + SQRT2 * maxr * i / 4
            txt_y = height / 2 + SQRT2 * maxr * i / 4
            self.canvas.create_text(
                txt_x, txt_y, text=dist, fill=FGCOL, font=self._font
            )

        for x, y, anc in (
            (width / 2, 5, N),
            (width / 2, height - 5, S),
            (5, height / 2, W),
            (width - 5, height / 2, E),
        ):
            self.canvas.create_text(
                x, y, text=anc.upper(), fill=FGCOL, font=self._font, anchor=anc
            )

    def _draw_stats(self, lbl_font: object):
        """
        Draw the stats in the corner of the plot.

        :param font lbl_font: Font to use.
        """

        if self._average is None:
            return

        self.canvas.delete(AVG)
        y = 5
        fh = self._fonth
        avg = f"Avg: {self._average.lat:.9f}, {self._average.lon:.9f}"
        self.canvas.create_text(
            5, y, text=avg, fill=PNTCOL, font=lbl_font, anchor=NW, tags=AVG
        )
        y += fh
        if self._stddev is not None:
            std = f"Std: {self._stddev.lat:.3e}, {self._stddev.lon:.3e}"
            self.canvas.create_text(
                5, y, text=std, fill=PNTCOL, font=lbl_font, anchor=NW, tags=AVG
            )
            y += fh
        np = len(self._points)
        pts = f"Pts: {np} {'!' if np >= self._maxpoints else ''}"
        self.canvas.create_text(
            5, y, text=pts, fill=PNTCOL, font=lbl_font, anchor=NW, tags=AVG
        )

    def _draw_point(self, position: Point, color: str = PNTCOL, size: int = 2):
        """
        Draw a point on the scatterplot.

        :param Point position: The point to draw
        :param str color: point color as string e.g. "orange"
        :param int size: size of circle (2)
        """

        if not point_in_bounds(self._bounds, position):
            return

        w, h = self.get_size()
        x, y = ll2xy(w, h, self._bounds, position)
        self.canvas.create_circle(x, y, size, fill=color, outline=color, tags=PNT)

    def _set_average(self):
        """
        Calculate the mean and standard deviation of all the lat/lon
        pairs visible on the scatter plot. Note that this will make
        for some weird results near poles.
        """

        lp = len(self._points)
        if HASSTATS and lp > 1:
            ave_lat = fmean(p.lat for p in self._points)
            ave_lon = fmean(p.lon for p in self._points)
            if not lp % STDINT:
                self._stddev = Point(
                    stdev(p.lat for p in self._points),
                    stdev(p.lon for p in self._points),
                )
        else:
            ave_lat = sum(p.lat for p in self._points) / lp
            ave_lon = sum(p.lon for p in self._points) / lp

        self._average = Point(ave_lat, ave_lon)

    def _set_bounds(self, center: Point):
        """
        Set bounding box of canvas baed on center point and
        plot range.

        :param Point center: centre of bounding box
        """

        cw, ch = self.get_size()
        disth = self._scale_factors[self._scale.get()]
        distw = self._scale_factors[self._scale.get()] * cw / ch
        t = get_point_at_vector(center, disth, 0)
        r = get_point_at_vector(center, distw, 90)
        b = get_point_at_vector(center, disth, 180)
        l = get_point_at_vector(center, distw, 270)
        self._bounds = Area(b.lat, l.lon, t.lat, r.lon)
        self._range = disth

        if self._bounds != self._lastbounds:
            self._init_frame()
            self._lastbounds = self._bounds

    def get_range_label(self) -> tuple:
        """
        Set range value and units according to magnitude.

        :return: range, units
        :rtype: tuple
        """

        if self._range >= 1000:
            rng = self._range / 1000
            unt = "km"
        elif self._range >= 1:
            rng = self._range
            unt = "m"
        else:
            rng = self._range * 100
            unt = "cm"
        return rng, unt

    def _redraw(self):
        """
        Redraw all the points on the scatter plot.
        """

        if not self._points:
            return

        self.canvas.delete(PNT)
        lp = len(self._points) - 1
        for i, pnt in enumerate(self._points):
            if i == lp:
                break
            self._draw_point(pnt, PNTCOL)
        if self._fixed is not None:
            self._draw_point(self._fixed, FIXCOL, 3)
        self._draw_point(self._points[-1], PNTTOPCOL)

        self._draw_stats(self._font)

    def update_frame(self):
        """
        Collect scatterplot data and update the plot.
        """

        self._updcount = (self._updcount + 1) % self._interval.get()
        if self._updcount:  # plot at every nth update
            return

        lat, lon = self.__app.gnss_status.lat, self.__app.gnss_status.lon
        try:
            lat = float(lat)
            lon = float(lon)
        except ValueError:
            return  # Invalid values for lat/lon get ignored.

        if lat == 0.0 and lon == 0.0:  # assume no fix
            return
        pos = Point(lat, lon)

        if self._points and pos == self._points[-1]:
            return  # Don't repeat exactly the last point.

        self._points.append(pos)
        if len(self._points) > self._maxpoints:
            self._cull_points()

        self._set_average()

        # set plot bounds based on range and center point
        middle = self._average
        try:
            self._fixed = Point(float(self._reflat.get()), float(self._reflon.get()))
            if self._centermode.get() == CTRFIX:
                middle = self._fixed
        except ValueError:
            self._fixed = None
            self._centermode.set(CTRAVG)
        self._set_bounds(middle)

        # update plotted point bounds
        self._minlat = min(lat, self._minlat)
        self._maxlat = max(lat, self._maxlat)
        self._minlon = min(lon, self._minlon)
        self._maxlon = max(lon, self._maxlon)
        if self._autorange.get():
            self._do_autorange(middle)

        self._redraw()

    def _cull_points(self):
        """
        Limit number of points in in-memory array.
        """

        if CULLMID:  # cull randomly from middle
            self._points.pop(randrange(1, len(self._points) - 100))
        else:  # cull from start
            self._points.pop(0)

    def _do_autorange(self, middle: Point):
        """
        Adjust range until all points in bounds.

        :param Point middle: center point of plot
        """

        out = True
        while out and self._scale.get() > 0:
            out = False

            # include fixed reference point in autorange
            if FIXINAUTO and not out and self._fixed is not None:
                if not point_in_bounds(self._bounds, self._fixed):
                    out = True
            if not out:
                if not point_in_bounds(
                    self._bounds, Point(self._minlat, self._minlon)
                ) or not point_in_bounds(
                    self._bounds, Point(self._maxlat, self._maxlon)
                ):
                    out = True
            if out:
                self._scale.set(self._scale.get() - 1)
                self._set_bounds(middle)

    def _on_resize(self, event):  # pylint: disable=unused-argument
        """
        Resize frame.

        :param Event event: resize event
        """

        self.width, self.height = self.get_size()
        self._font, self._fonth = scale_font(self.width, 10, 25, 20)
        self._init_frame()
        self._redraw()

    def get_size(self) -> tuple:
        """
        Get current canvas size.

        :return: window size (width, height)
        :rtype: tuple
        """

        self.update_idletasks()  # Make sure we know about resizing
        return self.canvas.winfo_width(), self.canvas.winfo_height()



"""
sysmon_frame.py

System Monitor frame for PyGPSClient application.

Shows cpu, memory, i/o status, core temperature and warning/error counts
for u-blox devices which support the MON-SYS and/or MON-COMMS UBX message
types.

Created on 30 Apr 2023

:author: semuadmin (Steve Smith)
:copyright: 2020 semuadmin
:license: BSD 3-Clause
"""

from tkinter import ALL, NW, Canvas, E, Frame, IntVar, N, Radiobutton, S, W

from pyubx2 import BOOTTYPE, UBXMessage

from pygpsclient.globals import BGCOL, FGCOL, PNTCOL, SYSMONVIEW, WIDGETU2
from pygpsclient.helpers import (
    bytes2unit,
    fontheight,
    hsv2rgb,
    scale_font,
    secs2unit,
    setubxrate,
)
from pygpsclient.strings import DLGENABLEMONSYS, DLGNOMONSYS, DLGWAITMONSYS, NA

MINFONT = 6  # minimum font size
MAXTEMP = 100  # °C
INSET = 4
SPACING = 5
DASH = (5, 2)
PORTIDS = {
    0x0000: "I2C",  # 0 I2C
    0x0100: "UART1",  # 256 UART1
    0x0101: "INT1",  # 257 inter-cpu connect
    0x0200: "INT2",  # 512 inter-cpu connect
    0x0201: "UART2",  # 513 UART2
    0x0300: "USB",  # 768 USB
    0x0400: "SPI",  # 1024 SPI
}
ACTIVE = ""
MAXLINES = 23
MINFONT = 4
MAXWAIT = 10


class SysmonFrame(Frame):
    """
    SysmonFrame class.
    """

    def __init__(self, app, *args, **kwargs):
        """
        Constructor.

        :param Frame app: reference to main tkinter application
        :param args: optional args to pass to Frame parent class
        :param kwargs: optional kwargs to pass to Frame parent class
        """

        self.__app = app  # Reference to main application class
        self.__master = self.__app.appmaster  # Reference to root class (Tk)

        Frame.__init__(self, self.__master, *args, **kwargs)

        def_w, def_h = WIDGETU2
        self.width = kwargs.get("width", def_w)
        self.height = kwargs.get("height", def_h)
        self._monsys_status = DLGENABLEMONSYS
        self._pending_confs = {}
        self._maxtemp = 0
        self._waits = 0
        self._mode = IntVar()
        self._mode.set(0)
        self._font = self.__app.font_sm
        self._fonth = fontheight(self._font)
        self._body()
        self._attach_events()

    def _body(self):
        """
        Set up frame and widgets.
        """

        self._can_sysmon = Canvas(self, width=self.width, height=self.height, bg=BGCOL)
        self._frm_status = Frame(self, bg=BGCOL)
        self._rad_actual = Radiobutton(
            self._frm_status,
            text="Actual I/O",
            variable=self._mode,
            value=0,
            fg=PNTCOL,
            bg=BGCOL,
            # selectcolor=BGCOL,
        )
        self._rad_pending = Radiobutton(
            self._frm_status,
            text="Pending I/O",
            variable=self._mode,
            value=1,
            fg=PNTCOL,
            bg=BGCOL,
            # selectcolor=BGCOL,
        )
        self._can_sysmon.grid(column=0, row=0, padx=0, pady=0, sticky=(N, S, W, E))
        self._frm_status.grid(column=0, row=1, padx=2, pady=2, sticky=(W, E))
        self._rad_actual.grid(column=0, row=0, padx=0, pady=0, sticky=W)
        self._rad_pending.grid(column=1, row=0, padx=0, pady=0, sticky=W)
        self.grid_columnconfigure(0, weight=1)
        self.grid_rowconfigure(0, weight=1)

    def _attach_events(self):
        """
        Bind events to frame.
        """

        self.bind("<Configure>", self._on_resize)
        self._can_sysmon.bind("<Double-Button-1>", self._on_clear)

    def init_chart(self):
        """
        Initialise sysmon chart.
        """

        self._can_sysmon.delete(ALL)
        self._can_sysmon.create_text(
            self.width / 2,
            self.height / 2,
            text=self._monsys_status,
            fill="orange",
        )

    def _on_clear(self, event):  # pylint: disable=unused-argument
        """
        Clear chart data and reinitialise canvas.

        :param Event event: clear event
        """

        self.__app.gnss_status.sysmon_data = {}
        self.__app.gnss_status.comms_data = {}
        self._maxtemp = 0
        self._monsys_status = DLGWAITMONSYS
        self.init_chart()

    def enable_messages(self, status: int):
        """
        Enable/disable UBX MON-SYS & MON-COMMS messages on
        default port(s).

        NB: CPU Load value only valid if rate = 1

        :param int status: 0 = off, 1 = on
        """

        for msgid in ("MON-SYS", "MON-COMMS"):
            setubxrate(self.__app, msgid, status)
        for msgid in ("ACK-ACK", "ACK-NAK"):
            self._set_pending(msgid, SYSMONVIEW)
        self._monsys_status = DLGWAITMONSYS
        self.init_chart()

    def _set_pending(self, msgid: int, ubxfrm: int):
        """
        Set pending confirmation flag for sysmon widget to
        signify that it's waiting for a confirmation message.

        :param int msgid: UBX message identity
        :param int ubxfrm: integer representing UBX configuration frame (0-13)
        """

        self._pending_confs[msgid] = ubxfrm

    def update_pending(self, msg: UBXMessage):
        """
        Receives polled confirmation message from the ubx_handler and
        updates sysmon status.

        :param UBXMessage msg: UBX config message
        """

        pending = self._pending_confs.get(msg.identity, False)
        if pending and msg.identity == "ACK-NAK":
            self._pending_confs.pop("ACK-NAK")
            self._monsys_status = DLGNOMONSYS

        if self._pending_confs.get("ACK-ACK", False):
            self._pending_confs.pop("ACK-ACK")

    def update_frame(self):
        """
        Plot MON-SPAN spectrum analysis.

        If no updates received after a given number of trys,
        assume receiver doesn't support MON-SYS or MON-COMMS.

        sysmon_data is a list of sys monitor parms.
        comms-data is a tuple of tx and rx values
        """

        sysdata = self.__app.gnss_status.sysmon_data
        commsdata = self.__app.gnss_status.comms_data

        if len(sysdata) + len(commsdata) == 0:
            self._waits += 1
            if self._waits >= MAXWAIT:
                self._monsys_status = DLGNOMONSYS
                self.init_chart()
            return

        self._monsys_status = ACTIVE
        self._waits = 0
        try:
            bootType = BOOTTYPE[sysdata.get("bootType", 0)]
            cpuLoad = sysdata.get("cpuLoad", NA)
            cpuLoadMax = sysdata.get("cpuLoadMax", NA)
            memUsage = sysdata.get("memUsage", NA)
            memUsageMax = sysdata.get("memUsageMax", NA)
            ioUsage = sysdata.get("ioUsage", NA)
            ioUsageMax = sysdata.get("ioUsageMax", NA)
            runTime = sysdata.get("runTime", NA)
            noticeCount = sysdata.get("noticeCount", NA)
            warnCount = sysdata.get("warnCount", NA)
            errorCount = sysdata.get("errorCount", NA)
            tempValue = sysdata.get("tempValue", NA)
            if isinstance(tempValue, int):
                tempValueP = tempValue * 100 / MAXTEMP
                self._maxtemp = max(tempValue, self._maxtemp) * 100 / MAXTEMP
            else:
                tempValueP = NA

            self.init_chart()
            y = self._fonth
            y = self._chart_parm(INSET, y, cpuLoadMax, cpuLoad, "CPU", "%")
            y = self._chart_parm(INSET, y, memUsageMax, memUsage, "Memory", "%")
            y = self._chart_parm(INSET, y, ioUsageMax, ioUsage, "I/O", "%")
            for port, pdata in sorted(commsdata.items()):
                y = self._chart_io(INSET, y, port, pdata)
            y += SPACING
            y = self._chart_parm(INSET, y, self._maxtemp, tempValueP, "Temp", "°C")

            rtm, rtmu = secs2unit(runTime)
            rtf = "" if rtmu == "secs" else ",.02f"
            txt = (
                f"Boot Type: {bootType}\n"
                + f"Runtime: {rtm:{rtf}} {rtmu}\n"
                + f"Notices: {noticeCount}, Warnings: {warnCount}, Errors: {errorCount}"
            )
            self._can_sysmon.create_text(
                INSET,
                y,
                text=txt,
                fill=FGCOL,
                anchor=NW,
                font=self._font,
            )
        except KeyError:  # invalid sysmon-data or comms-data
            self._monsys_status = DLGNOMONSYS
            self.init_chart()

    def _chart_parm(
        self,
        xoffset: int,
        y: int,
        maxval: int,
        val: int,
        lbl: str,
        unit: str,
    ) -> int:
        """
        Draw caption and current/max bar charts on canvas.
        """

        scale = (self.width - (3 * xoffset)) / 100
        x = xoffset
        self._can_sysmon.create_text(
            x,
            y,
            text=f"{lbl}: {val} {unit}",
            fill=FGCOL,
            anchor=W,
            font=self._font,
        )
        y += self._fonth
        if isinstance(maxval, (int, float)):
            self._can_sysmon.create_line(
                x,
                y,
                x + maxval * scale,
                y,
                fill=self._set_col(maxval),
                dash=DASH,
                width=self._fonth,
            )
        if isinstance(val, (int, float)):
            self._can_sysmon.create_line(
                x,
                y,
                x + val * scale,
                y,
                fill=self._set_col(val),
                width=self._fonth,
            )
            y += self._fonth + SPACING
        return y

    def _chart_io(self, xoffset: int, y: int, port: int, pdata: tuple):
        """
        Draw port I/O captions and tx/rx current/max bar charts on canvas.

        I/O byte counts will display actual or pending, depending
        on mode setting.

        :param int xoffset: x axis offset
        :param int y: y axis
        :param int port: port id
        :param tuple pdata: port data tx & rx
        :param int mode: 0 = total bytes, 1 = pending bytes
        """

        mod = self._mode.get()
        cap = self._font.measure("UART2 → 888.88 GB ← 888.88 GB: ⇄")
        scale = (self.width - cap - (3 * xoffset)) / 100
        x = xoffset
        txb, txbu = bytes2unit(pdata[3 if mod else 2])  # total or pending
        txf = "d" if txbu == "" else ".02f"
        rxb, rxbu = bytes2unit(pdata[7 if mod else 6])
        rxf = "d" if rxbu == "" else ".02f"
        txt = f"{PORTIDS.get(port, NA)} → {txb:{txf}} {txbu} ← {rxb:{rxf}} {rxbu}:"
        self._can_sysmon.create_text(  # port
            x,
            y,
            text=txt,
            fill=FGCOL,
            anchor=W,
            font=self._font,
        )
        self._can_sysmon.create_text(  # port
            x + cap - 1,
            y,
            text="⇄",
            fill=FGCOL,
            anchor=E,
            font=self._font,
        )
        p = -1
        for i in range(0, 8, 4):  # RX & TX
            self._can_sysmon.create_line(  # max
                x + cap,
                y + p,
                x + cap + pdata[i + 1] * scale,
                y + p,
                fill=self._set_col(pdata[i + 1]),
                dash=DASH,
                width=2,
            )
            self._can_sysmon.create_line(  # val
                x + cap,
                y + p,
                x + cap + pdata[i] * scale,
                y + p,
                fill=self._set_col(pdata[i]),
                width=2,
            )
            p += 4
        y += self._fonth
        return y

    def _set_col(self, val: float) -> str:
        """
        Set bar chart line color
        (green low, red high).

        :param val: value as %
        :return: color string
        :rtype: str
        """

        return hsv2rgb((100 - min(100, val)) / 300, 0.8, 0.8)

    def _on_resize(self, event):  # pylint: disable=unused-argument
        """
        Resize frame.

        :param event event: resize event
        """

        self.width, self.height = self.get_size()
        self._font, self._fonth = scale_font(self.width, 10, 35, 20)

    def get_size(self):
        """
        Get current canvas size.

        :return: window size (width, height)
        :rtype: tuple
        """

        self.update_idletasks()  # Make sure we know about any resizing
        return self._can_sysmon.winfo_width(), self._can_sysmon.winfo_height()



"""
graphview_frame.py

Graphview frame class for PyGPSClient application.

This handles a frame containing a graph of current satellite reception.

Created on 14 Sep 2020

:author: semuadmin (Steve Smith)
:copyright: 2020 semuadmin
:license: BSD 3-Clause
"""

from tkinter import ALL, BOTH, YES, Canvas, E, Frame

from pygpsclient.globals import (
    AXISCOL,
    BGCOL,
    FGCOL,
    GNSS_LIST,
    GRIDCOL,
    MAX_SNR,
    WIDGETU2,
)
from pygpsclient.helpers import fontheight, scale_font, snr2col

# Relative offsets of graph axes and legend
AXIS_XL = 19
AXIS_XR = 10
AXIS_Y = 22
OL_WID = 2
LEG_XOFF = AXIS_XL + 10
LEG_YOFF = 5
LEG_GAP = 5


class GraphviewFrame(Frame):
    """
    Graphview frame class.
    """

    def __init__(self, app, *args, **kwargs):
        """
        Constructor.

        :param Frame app: reference to main tkinter application
        :param args: optional args to pass to Frame parent class
        :param kwargs: optional kwargs to pass to Frame parent class
        """

        self.__app = app  # Reference to main application class
        self.__master = self.__app.appmaster  # Reference to root class (Tk)

        Frame.__init__(self, self.__master, *args, **kwargs)

        def_w, def_h = WIDGETU2
        self.width = kwargs.get("width", def_w)
        self.height = kwargs.get("height", def_h)
        self._font = self.__app.font_vsm
        self._fonth = fontheight(self._font)
        self._body()
        self._attach_events()

    def _body(self):
        """
        Set up frame and widgets.
        """

        self.grid_columnconfigure(0, weight=1)
        self.grid_rowconfigure(0, weight=1)
        self.can_graphview = Canvas(
            self, width=self.width, height=self.height, bg=BGCOL
        )
        self.can_graphview.pack(fill=BOTH, expand=YES)

    def _attach_events(self):
        """
        Bind events to frame.
        """

        self.bind("<Configure>", self._on_resize)
        self.can_graphview.bind("<Double-Button-1>", self._on_legend)

    def _on_legend(self, event):  # pylint: disable=unused-argument
        """
        On double-click - toggle legend on/off.

        :param event: event
        """

        self.__app.configuration.set(
            "legend_b", not self.__app.configuration.get("legend_b")
        )

    def init_frame(self):
        """
        Initialise graph view
        """

        w, h = self.width, self.height
        ticks = int(MAX_SNR / 10)
        self.can_graphview.delete(ALL)
        for i in range(ticks, 0, -1):
            y = (h - AXIS_Y) * i / ticks
            self.can_graphview.create_line(
                AXIS_XL, y, w - AXIS_XR + 2, y, fill=AXISCOL if i == ticks else GRIDCOL
            )
            self.can_graphview.create_text(
                10,
                y,
                text=str(MAX_SNR - (i * 10)),
                angle=90,
                fill=FGCOL,
                font=self._font,
            )
        self.can_graphview.create_line(AXIS_XL, 5, AXIS_XL, h - AXIS_Y, fill=AXISCOL)
        self.can_graphview.create_line(
            w - AXIS_XR + 2, 5, w - AXIS_XR + 2, h - AXIS_Y, fill=AXISCOL
        )
        self.can_graphview.create_text(
            AXIS_XR,
            5,
            text="CN₀ dB",
            angle=90,
            fill=FGCOL,
            anchor=E,
            font=self._font,
        )

        if self.__app.configuration.get("legend_b"):
            self._draw_legend()

    def _draw_legend(self):
        """
        Draw GNSS color code legend
        """

        w = self.width / 10
        h = self.height / 15

        for i, (_, (gnssName, gnssCol)) in enumerate(GNSS_LIST.items()):
            x = LEG_XOFF + w * i
            self.can_graphview.create_rectangle(
                x,
                LEG_YOFF,
                x + w - LEG_GAP,
                LEG_YOFF + h,
                outline=gnssCol,
                fill=BGCOL,
                width=OL_WID,
            )
            self.can_graphview.create_text(
                (x + x + w - LEG_GAP) / 2,
                LEG_YOFF + h / 2,
                text=gnssName,
                fill=FGCOL,
                font=self._font,
            )

    def update_frame(self):
        """
        Plot satellites' signal-to-noise ratio (cno).
        Automatically adjust y axis according to number of satellites in view.
        """

        data = self.__app.gnss_status.gsv_data
        siv = len(self.__app.gnss_status.gsv_data)

        if siv == 0:
            return

        w, h = self.width, self.height
        self.init_frame()

        offset = AXIS_XL + 2
        colwidth = (w - AXIS_XL - AXIS_XR + 1) / siv
        # scale x axis label according to siv
        svfont, _ = scale_font(self.width, 6, siv, 14)
        for d in sorted(data.values()):  # sort by ascending gnssid, svid
            gnssId, prn, _, _, snr = d
            if snr in ("", "0", 0):
                snr = 1  # show 'place marker' in graph
            else:
                snr = int(snr)
            snr_y = int(snr) * (h - AXIS_Y - 1) / MAX_SNR
            (_, ol_col) = GNSS_LIST[gnssId]
            prn = f"{int(prn):02}"
            self.can_graphview.create_rectangle(
                offset,
                h - AXIS_Y - 1,
                offset + colwidth - OL_WID,
                h - AXIS_Y - 1 - snr_y,
                outline=ol_col,
                fill=snr2col(snr),
                width=OL_WID,
            )
            self.can_graphview.create_text(
                offset + colwidth / 2,
                h - 10,
                text=prn,
                fill=FGCOL,
                font=svfont,
                angle=35,
            )
            offset += colwidth

        self.can_graphview.update_idletasks()

    def _on_resize(self, event):  # pylint: disable=unused-argument
        """
        Resize frame

        :param event event: resize event
        """

        self.width, self.height = self.get_size()
        self._font, self._fonth = scale_font(self.width, 8, 25, 16)

    def get_size(self):
        """
        Get current canvas size.

        :return: window size (width, height)
        :rtype: tuple
        """

        self.update_idletasks()  # Make sure we know about any resizing
        return self.can_graphview.winfo_width(), self.can_graphview.winfo_height()



"""
sqlite_handler.py

sqlite3 database class for PyGPSClient application.

This handles all the sqlite3 database updates.

**NB**: Functionality is subject to the following Python
environmental criteria:

1. The Python environment must support the loading of
   sqlite3 extensions i.e. it must have been compiled
   with the `--enable-loadable-sqlite-extensions` option.
2. The mod_spatialite module (.so, .dll or .dylib)
   must be installed and in the `PATH`/`LD_LIBRARY_PATH`.

Created on 13 Sep 2025

:author: semuadmin (Steve Smith)
:copyright: 2020 semuadmin
:license: BSD 3-Clause
"""

import logging
import sqlite3
import traceback
from datetime import datetime, timezone
from os import path

from pynmeagps import ecef2llh

from pygpsclient.globals import ERRCOL, HOME, INFOCOL, OKCOL
from pygpsclient.helpers import makeval
from pygpsclient.strings import NA

# path to mod_spatialite module, if required
# SLPATH = "C:/Program Files/QGIS 3.44.1/bin"
# environ["PATH"] = SLPATH + ";" + environ["PATH"]

SQLVER = sqlite3.sqlite_version
"""sqlite3 version"""
SQLOK = SQLVER
"""No SQL error"""
SQLERR = "SQL error"
"""Non-specific SQL error"""
NOEXT = "No ext"
"""sqlite3 extensions not supported"""
NOMODS = "No m_s"
"""sqlite3 mod_spatialite extension not found"""
DBNAME = "pygpsclient.sqlite"
"""Default database name"""
DBINMEM = ":memory:"
"""In-memory database name"""
TBNAME = "pygpsclient"
"""Default table name"""
SQLBEGIN = "BEGIN TRANSACTION;"
SQLCOMMIT = "COMMIT;"

SQLC1 = (
    SQLBEGIN
    + (
        "DROP TABLE IF EXISTS {table};"
        "CREATE TABLE {table} (id INTEGER PRIMARY KEY, utc REAL, fix TEXT, hae REAL, "
        "speed REAL, track REAL, siv INTEGER, sip INTEGER, pdop REAL, hdop REAL, vdop REAL, "
        "hacc REAL, vacc REAL, diffcorr INTEGER, diffage INTEGER, diffstat TEXT, "
        "baselon REAL, baselat REAL, basehae REAL);"
        "SELECT AddGeometryColumn('{table}', 'geom', 4326, 'POINT', 'XYZ');"
        "SELECT CreateSpatialIndex('{table}', 'geom');"
    )
    + SQLCOMMIT
)
"""SQL for creating database and table with lat/lon/hmsl as 3D POINTZ"""

SQLI3D = (
    "INSERT INTO {table} (geom, utc, fix, hae, speed, track, siv, sip, pdop, "
    "hdop, vdop, hacc, vacc, diffcorr, diffage, diffstat, baselon, baselat, basehae) "
    "VALUES (GeomFromText('POINTZ({lon} {lat} {hmsl})', 4326), {utc}, '{fix}', "
    "{hae}, {speed}, {track}, {siv}, {sip}, {pdop}, {hdop}, {vdop}, {hacc}, "
    "{vacc}, {diffcorr}, {diffage}, '{diffstat}', {baselon}, {baselat}, {basehae});"
)
"""SQL for inserting row into table"""

SQLSEL = (
    "SELECT ST_X(geom), ST_Y(geom), ST_Z(geom), utc, fix, hae, speed, track, siv, "
    "sip, pdop, hdop, vdop, hacc, vacc, diffcorr, diffage, diffstat, baselon, "
    "baselat, basehae from {table} ORDER BY utc LIMIT {limit} OFFSET {offset}"
)
"""SQL for retrieving rows from table"""


class SqliteHandler:
    """
    sqlite handler class.
    """

    def __init__(self, app):
        """
        Constructor.

        :param Frame app: reference to main tkinter application

        """

        self.__app = app  # Reference to main application class
        self.__master = self.__app.appmaster  # Reference to root class (Tk)
        self.logger = logging.getLogger(__name__)

        self._db = None
        self._table = None
        self._dbname = None
        self._connection = None
        self._cursor = None

    def _create(
        self,
        tbname: str = TBNAME,
    ) -> int:
        """
        Create sqlite3 connection and cursor.

        :param str dbpath: path to sqlite3 database file
        :param str dbname: name of sqlite3 database file
        :param str tbname: name of table containing gnss data
        :return: return code
        :rtype: int
        """

        try:
            self.__app.set_status(
                f"Database {self._db} initialising - please wait...", INFOCOL
            )
            self.logger.debug("Spatial metadata initialisation in progress...")
            self._connection.execute("SELECT InitSpatialMetaData();")
            self.logger.debug("Spatial metadata initialisation complete")
            self._cursor = self._connection.cursor()
            self._cursor.executescript(SQLC1.format(table=tbname))
            return SQLOK
        except sqlite3.Error as err:
            self.__app.set_status(f"Error initialising spatial database {err}", ERRCOL)
            self.logger.debug(traceback.format_exc())
            return SQLERR

    def open(
        self,
        dbpath: str = HOME,
        dbname: str = DBNAME,
        tbname: str = TBNAME,
    ) -> str:
        """
        Create sqlite3 connection and cursor.

        :param str dbpath: path to sqlite3 database file
        :param str dbname: name of sqlite3 database file
        :param str tbname: name of table containing gnss data
        :return: result
        :rtype: str
        """

        try:
            db = ""
            if dbname == DBINMEM:  # check for spatial support
                db = dbname
                exists = True
            else:
                db = path.join(dbpath, dbname)
                exists = path.exists(db)
                self._db = db
                self._table = tbname
                self._dbname = dbname
            self._connection = sqlite3.connect(db)
            self._connection.enable_load_extension(True)
            self._connection.load_extension("mod_spatialite")
            if not exists:
                if not self._create(tbname):
                    return 0
            self._cursor = self._connection.cursor()
            if dbname == DBINMEM:
                self._connection.close()
            else:
                self.__app.set_status(f"Database {self._db} opened", OKCOL)
            return SQLOK
        except AttributeError as err:
            self.__app.set_status(f"SQL error: {err}", ERRCOL)
            self.logger.debug(traceback.format_exc())
            return NOEXT  # extensions not supported
        except sqlite3.OperationalError as err:
            self.__app.set_status(f"SQL error {db}: {err}", ERRCOL)
            self.logger.debug(traceback.format_exc())
            return NOMODS  # no mod_spatial extension found
        except sqlite3.Error as err:
            self.__app.set_status(f"SQL error {db}: {err}", ERRCOL)
            self.logger.debug(traceback.format_exc())
            return SQLERR  # other sqlite error

    def close(self):
        """
        Close database connection.
        """

        if self._connection is not None:
            try:
                self._connection.cursor()
            except sqlite3.Error:
                return
            self._connection.close()

    def load_data(self, ignore_null: bool = True) -> str:
        """
        Load current gnss data (from `self.__app.gnss_status`) into database.

        :param bool ignore_null: ignore null position flag
        :return: result
        :rtype: str
        """

        gnss = self.__app.gnss_status
        if ignore_null and gnss.lat == 0.0 and gnss.lon == 0.0:
            self.logger.debug("Ignored null lat/lon value")
            return SQLOK

        try:
            baselat, baselon, basehae = ecef2llh(
                gnss.base_ecefx, gnss.base_ecefy, gnss.base_ecefz
            )
            basehae = 0.0 if basehae == -10000000.0 else basehae
            utc = datetime.combine(datetime.now(timezone.utc).date(), gnss.utc)
            sql = SQLI3D.format(
                table=self._table,
                lon=makeval(gnss.lon),
                lat=makeval(gnss.lat),
                hmsl=makeval(gnss.alt),
                utc=float(utc.strftime("%Y%m%d%H%M%S.%f")),
                fix=gnss.fix,
                hae=makeval(gnss.hae),
                speed=makeval(gnss.speed),
                track=makeval(gnss.track),
                siv=makeval(gnss.siv, 0),
                sip=makeval(gnss.sip, 0),
                pdop=makeval(gnss.pdop),
                hdop=makeval(gnss.hdop),
                vdop=makeval(gnss.vdop),
                hacc=makeval(gnss.hacc),
                vacc=makeval(gnss.vacc),
                diffcorr=makeval(gnss.diff_corr),
                diffage=makeval(gnss.diff_age, 0),
                diffstat=makeval(gnss.diff_station, NA),
                baselon=makeval(baselon),
                baselat=makeval(baselat),
                basehae=makeval(basehae),
            )
            sql = SQLBEGIN + sql + SQLCOMMIT
            self._cursor.executescript(sql)
            self.logger.debug(f"Executed SQL statement {sql}")
            return SQLOK
        except sqlite3.Error as err:
            self.__app.set_status(f"SQL error: {err}", ERRCOL)
            self.logger.debug(traceback.format_exc())
            return SQLERR

    def get_data(self, limit: int = 100, offset: int = 0) -> object:
        """
        Retrieve rows from table as iterable, ordered by utc timestamp.

        Check for SQLERR before iterating results.

        :param int limit: number of rows
        :param int offset: starting row
        :return: iterable results
        :rtype: results or error code as string
        """

        try:
            self._cursor.execute(
                SQLSEL.format(table=self._table, limit=limit, offset=offset)
            )
            return self._cursor.fetchall()
        except sqlite3.Error:
            self.logger.debug(traceback.format_exc())
            return SQLERR

    @property
    def database(self) -> str:
        """
        Getter for database name.

        :return: database path/name
        :rtype: str
        """

        return self._db



"""
spectrum_frame.py

Spectrum Analysis frame class for PyGPSClient application.

This handles a frame containing a spectrum analysis chart from
a MON-SPAN message.

Created on 23 Dec 2022

:author: semuadmin (Steve Smith)
:copyright: 2020 semuadmin
:license: BSD 3-Clause
"""

import logging
from tkinter import ALL, NW, Canvas, Checkbutton, E, Frame, IntVar, N, S, W

from pyubx2 import UBXMessage

from pygpsclient.globals import (
    BGCOL,
    FGCOL,
    GNSS_LIST,
    GRIDCOL,
    PNTCOL,
    SPECTRUMVIEW,
    WIDGETU2,
    AreaXY,
)
from pygpsclient.helpers import data2xy, fontheight, scale_font, setubxrate, xy2data
from pygpsclient.strings import DLGENABLEMONSPAN, DLGNOMONSPAN, DLGWAITMONSPAN

# Graph dimensions
RESFONT = 24  # font size relative to widget size
MINFONT = 7  # minimum font size
OL_WID = 1
MIN_DB = 0
MAX_DB = 200
MIN_HZ = 1130000000
MAX_HZ = 1650000000
TICK_DB = 20  # 20 dB divisions
TICK_GHZ = 40000000  # 40 MHz divisions
RF_BANDS = {
    "B1": 1575420000,
    "B3": 1268520000,
    "B2": 1202025000,
    "B2a": 1176450000,
    "E6": 1278750000,
    "E5b": 1202025000,
    "E5a": 1176450000,
    "E1": 1575420000,
    "G3": 1202025000,
    "G2": 1248060000,
    "G1": 1600995000,
    "L5": 1176450000,
    "L2": 1227600000,
    "L1": 1575420000,
}
RF_LIST = {
    0: "aquamarine2",
    1: "orange",
    2: "deepskyblue",
    3: "sandybrown",
    4: "dodgerblue",
    5: "darkorange",
}
RF_LIST_SNAPSHOT = {
    0: "gray75",
    1: "gray60",
    2: "gray65",
    3: "gray50",
    4: "gray55",
    5: "gray40",
}
ACTIVE = ""
MODEINIT = "init"
MODELIVE = "live"
MODESNAP = "snap"


class SpectrumviewFrame(Frame):
    """
    Spectrumview frame class.
    """

    def __init__(self, app, *args, **kwargs):
        """
        Constructor.

        :param Frame app: reference to main tkinter application
        :param args: optional args to pass to Frame parent class
        :param kwargs: optional kwargs to pass to Frame parent class
        """

        self.__app = app  # Reference to main application class
        self.__master = self.__app.appmaster  # Reference to root class (Tk)
        self.logger = logging.getLogger(__name__)

        Frame.__init__(self, self.__master, *args, **kwargs)

        def_w, def_h = WIDGETU2
        self.width = kwargs.get("width", def_w)
        self.height = kwargs.get("height", def_h)
        self._mindb = MIN_DB
        self._maxdb = MAX_DB
        self._minhz = MIN_HZ
        self._maxhz = MAX_HZ
        self._monspan_status = DLGENABLEMONSPAN
        self._pending_confs = {}
        self._showrf = True
        self._chartpos = None
        self._spectrum_snapshot = []
        self._pgaoffset = IntVar()
        self._font = self.__app.font_sm
        self._fonth = fontheight(self._font)
        self._body()
        self._attach_events()

    def _body(self):
        """
        Set up frame and widgets.
        """

        self.grid_columnconfigure(0, weight=1)
        self.grid_rowconfigure(0, weight=1)
        self.grid_rowconfigure(1, weight=0)
        self.can_spectrumview = Canvas(
            self, width=self.width, height=self.height, bg=BGCOL
        )
        self.chk_pgaoffset = Checkbutton(
            self,
            text="PGA Offset",
            fg=PNTCOL,
            bg=BGCOL,
            variable=self._pgaoffset,
            anchor=W,
        )
        # self.can_spectrumview.pack(fill=BOTH, expand=YES)
        self.can_spectrumview.grid(column=0, row=0, columnspan=3, sticky=(N, S, E, W))
        self.chk_pgaoffset.grid(column=0, row=1, sticky=(W, E))

    def _attach_events(self):
        """
        Bind events to frame.
        """

        self.bind("<Configure>", self._on_resize)
        self.can_spectrumview.bind("<Button-1>", self._on_click)
        self.can_spectrumview.bind("<Double-Button-1>", self._on_toggle_rf)
        self.can_spectrumview.bind("<Button-2>", self._on_snapshot)
        self.can_spectrumview.bind("<Double-Button-2>", self._on_clear_snapshot)

    def init_frame(self):
        """
        Initialise spectrum chart.
        """

        offset = self._fonth + 4
        w = self.width - offset * 2
        h = self.height - offset
        bounds = AreaXY(self._minhz, self._mindb, self._maxhz, self._maxdb)
        self.can_spectrumview.delete(ALL)

        # plot y (dB) axis grid
        i = 0
        for db in range(self._mindb, self._maxdb, TICK_DB):
            x1, y1 = data2xy(w, h, bounds, self._minhz, db, offset)
            x2, y2 = data2xy(w, h, bounds, self._maxhz, db, offset)
            self.can_spectrumview.create_line(
                x1,
                y1,
                x2 + 1,
                y1,
                fill=GRIDCOL if i else FGCOL,
                tags=MODEINIT,
            )
            self.can_spectrumview.create_text(
                x1,
                y1,
                text=f"{db}",
                angle=90,
                fill=FGCOL,
                font=self._font,
                anchor=S,
                tags=MODEINIT,
            )
            i += 1

        # plot x (Hz) axis grid
        i = 0
        for hz in range(self._minhz, self._maxhz, TICK_GHZ):
            x1, y1 = data2xy(w, h, bounds, hz, self._mindb, offset)
            x2, y2 = data2xy(w, h, bounds, hz, self._maxdb, offset)
            self.can_spectrumview.create_line(
                x1,
                y1 - 1,
                x1,
                y2,
                fill=GRIDCOL if i else FGCOL,
                tags=MODEINIT,
            )
            self.can_spectrumview.create_text(
                x1,
                y1,
                text=f"{hz / 1e9:.2f}",  # GHz
                fill=FGCOL,
                font=self._font,
                anchor=N,
                tags=MODEINIT,
            )
            i += 1

        x, y = data2xy(w, h, bounds, self._maxhz, self._mindb, offset)
        self.can_spectrumview.create_text(
            x,
            y,
            text="GHz",
            fill=FGCOL,
            font=self._font,
            anchor="se",
            tags=MODEINIT,
        )
        x, y = data2xy(w, h, bounds, self._minhz + self._fonth, self._maxdb - 5, offset)
        self.can_spectrumview.create_text(
            x,
            y,
            text="dB",
            fill=FGCOL,
            angle=90,
            font=self._font,
            anchor="ne",
            tags=MODEINIT,
        )

        # display 'enable MON-SPAN' warning
        self.can_spectrumview.create_text(
            self.width / 2,
            self.height / 2,
            text=self._monspan_status,
            fill="orange",
            tags=MODEINIT,
        )

    def reset(self):
        """
        Reset spectrumview frame.
        """

        self.__app.gnss_status.spectrum_data = []
        self._chartpos = None
        self._pgaoffset.set(0)
        self.can_spectrumview.delete(ALL)
        self.update_frame()

    def enable_messages(self, status: bool):
        """
        Enable/disable UBX MON-SPAN message.

        :param bool status: 0 = off, 1 = on
        """

        setubxrate(self.__app, "MON-SPAN", status)
        for msgid in ("ACK-ACK", "ACK-NAK"):
            self._set_pending(msgid, SPECTRUMVIEW)
        self._monspan_status = DLGWAITMONSPAN

    def _set_pending(self, msgid: int, ubxfrm: int):
        """
        Set pending confirmation flag for Spectrumview frame to
        signify that it's waiting for a confirmation message.

        :param int msgid: UBX message identity
        :param int ubxfrm: integer representing UBX configuration frame (0-6)
        """

        self._pending_confs[msgid] = ubxfrm

    def update_pending(self, msg: UBXMessage):
        """
        Receives polled confirmation message from the ubx_handler and
        updates spectrumview canvas.

        :param UBXMessage msg: UBX config message
        """

        pending = self._pending_confs.get(msg.identity, False)

        if pending and msg.identity == "ACK-NAK":
            self.reset()
            w, h = self.width, self.height
            self.can_spectrumview.create_text(
                w / 2,
                h / 2,
                text=DLGNOMONSPAN,
                fill="orange",
                anchor=S,
            )
            self._pending_confs.pop("ACK-NAK")
            self._monspan_status = DLGNOMONSPAN
            self.init_frame()

        if self._pending_confs.get("ACK-ACK", False):
            self._pending_confs.pop("ACK-ACK")

    def update_frame(self):
        """
        Plot MON-SPAN spectrum analysis.

        spectrum_data is list of tuples (spec, spn, res, ctr, pga),
        one item per RF block.
        """

        rfblocks = self.__app.gnss_status.spectrum_data
        if len(rfblocks) == 0:
            return
        self._monspan_status = ACTIVE
        self._update_plot(rfblocks)

        if self._spectrum_snapshot != []:
            self._update_plot(self._spectrum_snapshot, MODESNAP, RF_LIST_SNAPSHOT)

    def _update_plot(self, rfblocks: list, mode: str = MODELIVE, colors: dict = None):
        """
        Update spectrum plot with live or snapshot rf block data.

        :param list rfblocks: array of spectrum rf blocks
        :param dict colors: dictionary of color for each rf block
        :param str mode: plot mode ("live" or "snap"shot)
        """

        offset = self._fonth + 4
        w = self.width - offset * 2
        h = self.height - offset

        if colors is None:
            colors = RF_LIST

        self._mindb = MIN_DB
        self._maxdb = MAX_DB
        if self._pgaoffset.get():
            self._maxdb += 40

        specxy, self._minhz, self._maxhz = self._get_limits(rfblocks)
        bounds = AreaXY(self._minhz, self._mindb, self._maxhz, self._maxdb)
        if mode == MODESNAP:
            self.can_spectrumview.delete(MODESNAP)
        else:
            self.init_frame()
            # plot frequency band markers
            if self._showrf:
                self._plot_rf_markers(w, h, bounds, offset, mode)

        # for each RF block in MON-SPAN message
        for i, rfblock in enumerate(specxy):
            col = colors[i % len(colors)]

            # draw legend for this RF block
            x1, y1 = (50 * (i + 1) + (i * self._fonth), self._fonth * 2)
            x2, y2 = (50 + 50 * (i + 1) + (i * self._fonth), self._fonth * 2)
            if mode == MODESNAP:
                y1 += self._fonth
                y2 += self._fonth
            else:
                self.can_spectrumview.create_text(
                    (x1 + x2) / 2,
                    y1,
                    text=f"RF {i + 1}",
                    fill=FGCOL,
                    font=self._font,
                    anchor=S,
                    tags=mode,
                )
            self.can_spectrumview.create_line(
                x1,
                y1,
                x2,
                y1,
                fill=col,
                width=OL_WID,
                tags=mode,
            )

            # plot spectrum for this RF block
            x2 = 5
            y2 = self.height - 5
            for n, (hz, db) in enumerate(rfblock):
                x1, y1 = x2, y2
                x2, y2 = data2xy(w, h, bounds, hz, db, offset)
                if n:
                    self.can_spectrumview.create_line(
                        x1,
                        y1,
                        x2,
                        y2,
                        fill=col,
                        width=OL_WID,
                        tags=mode,
                    )

        # display any flagged chart position
        if self._chartpos is not None:
            x, y, hz, db = self._chartpos
            self.can_spectrumview.create_text(
                x,
                y,
                text=f"{hz:.3f} GHz\n{db:.1f} dB",
                fill=FGCOL,
                font=self._font,
                anchor="center",
                tags=mode,
            )

    def _plot_rf_markers(self, w: int, h: int, bounds: AreaXY, offset: int, mode: int):
        """
        Plot RF band markers

        :param int w: plots width
        :param int h: plot height
        :param AreaXY bounds: data bounds
        :param offset: _description_
        """
        # pylint: disable=too-many-arguments, too-many-positional-arguments

        for nam, frq in RF_BANDS.items():
            if self._minhz < frq < self._maxhz:
                x1, y1 = data2xy(w, h, bounds, frq, self._maxdb, offset)
                x2, y2 = data2xy(w, h, bounds, frq, self._mindb, offset)
                yoff, col = {
                    "L": (self._fonth, GNSS_LIST[0][1]),  # GPS
                    "G": (self._fonth * 2, GNSS_LIST[6][1]),  # GLONASS
                    "E": (self._fonth * 3, GNSS_LIST[2][1]),  # Galileo
                    "S": (self._fonth * 3, GNSS_LIST[2][1]),  # Galileo SAR
                    "B": (self._fonth * 4, GNSS_LIST[3][1]),  # Beidou
                }[nam[0:1]]
                if nam not in (
                    "E1",
                    "E5a",
                    "E5b",
                    "B2a",
                    "B2",
                    "B1",
                ):  # same freq as other bands
                    self.can_spectrumview.create_line(
                        x1,
                        y1,
                        x1,
                        y2,
                        fill=col,
                        dash=(5, 2),
                        width=OL_WID,
                        tags=mode,
                    )
                self.can_spectrumview.create_text(
                    x2 + 2,
                    y2 - yoff - 1,
                    text=nam,
                    fill=col,
                    anchor=NW,
                    font=self._font,
                    tags=mode,
                )

    def _get_limits(self, rfblocks: list) -> tuple:
        """
        Get axis limits for all RF blocks and convert
        spectrum arrays to (x,y) arrays.
        Frequencies expressed as GHz.

        :param list rfblocks: RF Blocks
        :return: tuple of points and axis limits
        :rtype: tuple
        """

        minhz = 999 * 1e9
        maxhz = 0
        specxy = []

        # for each RF block in MON-SPAN message
        for i, rfblock in enumerate(rfblocks):
            (spec, spn, res, ctr, pga) = rfblock
            minhz = int(min(minhz, ctr - res * (spn / res) / 2))
            maxhz = int(max(maxhz, ctr + res * (spn / res) / 2))
            spanhz = []
            for i, db in enumerate(spec):
                if self._pgaoffset.get():
                    db += pga  # compensate for programmable gain
                hz = int(ctr - (spn / 2) + (res * i))
                spanhz.append((hz, db))
            specxy.append(spanhz)

        return (
            specxy,
            int(minhz - TICK_GHZ / 2),
            int(maxhz + TICK_GHZ / 2),
        )

    def _on_click(self, event):
        """
        Save flagged chart position.
        """

        offset = self._fonth + 4
        w = self.width - offset * 2
        h = self.height - offset
        bounds = AreaXY(self._minhz, self._mindb, self._maxhz, self._maxdb)
        hz, db = xy2data(w, h, bounds, event.x - offset, event.y)
        self._chartpos = (event.x, event.y, hz / 1e9, db)

    def _on_toggle_rf(self, event):  # pylint: disable=unused-argument
        """
        Toggle RF band markers on/off.
        """

        self._showrf = not self._showrf
        self._chartpos = None

    def _on_snapshot(self, event):  # pylint: disable=unused-argument
        """
        Capture snapshot of current spectrum.
        """

        self._spectrum_snapshot = self.__app.gnss_status.spectrum_data

    def _on_clear_snapshot(self, event):  # pylint: disable=unused-argument
        """
        Clear snapshot of current spectrum.
        """

        self._spectrum_snapshot = []
        self.can_spectrumview.delete("snap")

    def _on_resize(self, event):  # pylint: disable=unused-argument
        """
        Resize frame.

        :param event event: resize event
        """

        self.width, self.height = self.get_size()
        self._font, self._fonth = scale_font(self.width, 8, 25, 20)
        self._chartpos = None

    def get_size(self):
        """
        Get current canvas size.

        :return: window size (width, height)
        :rtype: tuple
        """

        self.update_idletasks()  # Make sure we know about any resizing
        return self.can_spectrumview.winfo_width(), self.can_spectrumview.winfo_height()



"""
confirm_box.py

Confirm action dialog class.
Provides better consistency across different OS platforms
than using messagebox.askyesno()

Created on 17 Apr 2021

:author: semuadmin (Steve Smith)
:copyright: 2020 semuadmin
:license: BSD 3-Clause

"""

from tkinter import Button, Label, Toplevel, W


class ConfirmBox(Toplevel):
    """
    Confirm action dialog class.
    Provides better consistency across different OS platforms
    than using messagebox.askyesno()

    Returns True if OK, False if Cancel
    """

    def __init__(self, parent, title, prompt):
        """
        Constructor

        :param parent: parent dialog
        :param string title: title
        :param string prompt: prompt to be displayed
        """

        self.__master = parent
        Toplevel.__init__(self, parent)
        self.title(title)  # pylint: disable=E1102
        self.resizable(False, False)
        Label(self, text=prompt, anchor=W).grid(
            row=0, column=0, columnspan=2, padx=3, pady=5
        )
        Button(self, command=self._on_ok, text="OK", width=8).grid(
            row=1, column=0, padx=3, pady=3
        )
        Button(self, command=self._on_cancel, text="Cancel", width=8).grid(
            row=1, column=1, padx=3, pady=3
        )
        self.lift()  # Put on top of
        self.grab_set()  # Make modal
        self._rc = False

        self._centre()

    def _on_ok(self, event=None):  # pylint: disable=unused-argument
        """
        OK button handler
        """

        self._rc = True
        self.destroy()

    def _on_cancel(self, event=None):  # pylint: disable=unused-argument
        """
        Cancel button handler
        """

        self._rc = False
        self.destroy()

    def _centre(self):
        """
        Centre dialog in parent
        """

        # self.update_idletasks()
        dw = self.winfo_width()
        dh = self.winfo_height()
        mx = self.__master.winfo_x()
        my = self.__master.winfo_y()
        mw = self.__master.winfo_width()
        mh = self.__master.winfo_height()
        self.geometry(f"+{int(mx + (mw/2 - dw/2))}+{int(my + (mh/2 - dh/2))}")

    def show(self):
        """
        Show dialog

        :return: True (OK) or False (Cancel)
        :rtype: bool
        """

        self.wm_deiconify()
        self.wait_window()
        return self._rc



"""
toplevel_dialog.py

Top Level container dialog which displays child frames
within a scrollable and resizeable canvas, primarily
to allow dialog to be usable on low resolution screens.

Created on 19 Sep 2020

:author: semuadmin (Steve Smith)
:copyright: 2020 semuadmin
:license: BSD 3-Clause
"""

from tkinter import (
    ALL,
    HORIZONTAL,
    NW,
    VERTICAL,
    Button,
    Canvas,
    E,
    Frame,
    Label,
    N,
    S,
    Scrollbar,
    StringVar,
    Toplevel,
    W,
)

from PIL import Image, ImageTk

from pygpsclient.globals import (
    ERRCOL,
    ICON_BLANK,
    ICON_CONFIRMED,
    ICON_CONN,
    ICON_DISCONN,
    ICON_END,
    ICON_EXIT,
    ICON_LOAD,
    ICON_PENDING,
    ICON_REDRAW,
    ICON_SEND,
    ICON_START,
    ICON_WARNING,
    MINHEIGHT,
    MINWIDTH,
    RESIZE,
)
from pygpsclient.helpers import check_lowres


class ToplevelDialog(Toplevel):
    """
    ToplevelDialog class.
    """

    def __init__(self, app, dlgname: str, dim: tuple = (MINHEIGHT, MINWIDTH)):
        """
        Constructor.

        :param Frame app: reference to main tkinter application
        :param str dlgname: dialog name
        :param tuple dim: initial dimensions (height, width)
        """

        self.__app = app  # Reference to main application class
        self.__master = self.__app.appmaster  # Reference to root class (Tk)
        self._dlgname = dlgname
        self.lowres, (self.height, self.width) = check_lowres(self.__master, dim)

        super().__init__()

        if self.__app.configuration.get("transient_dialog_b"):
            self.transient(self.__app)
        self.title(dlgname)  # pylint: disable=E1102
        if self.__app.dialog_state.state[dlgname][RESIZE]:
            self.resizable(True, True)
        else:
            self.resizable(self.lowres, self.lowres)
        self.protocol("WM_DELETE_WINDOW", self.on_exit)
        self.img_none = ImageTk.PhotoImage(Image.open(ICON_BLANK))
        self.img_confirmed = ImageTk.PhotoImage(Image.open(ICON_CONFIRMED))
        self.img_conn = ImageTk.PhotoImage(Image.open(ICON_CONN))
        self.img_disconn = ImageTk.PhotoImage(Image.open(ICON_DISCONN))
        self.img_end = ImageTk.PhotoImage(Image.open(ICON_END))
        self.img_exit = ImageTk.PhotoImage(Image.open(ICON_EXIT))
        self.img_load = ImageTk.PhotoImage(Image.open(ICON_LOAD))
        self.img_pending = ImageTk.PhotoImage(Image.open(ICON_PENDING))
        self.img_redraw = ImageTk.PhotoImage(Image.open(ICON_REDRAW))
        self.img_send = ImageTk.PhotoImage(Image.open(ICON_SEND))
        self.img_start = ImageTk.PhotoImage(Image.open(ICON_START))
        self.img_warn = ImageTk.PhotoImage(Image.open(ICON_WARNING))
        self._status = StringVar()

        self._con_body()

    def on_expand(self):
        """
        Automatically expand container canvas when sub-frames are resized.
        """

        self._can_container.event_generate("<Configure>")

    def _con_body(self):
        """
        Set up scrollable frame and widgets.
        """

        # create container frame
        if self.lowres:
            x_scrollbar = Scrollbar(self, orient=HORIZONTAL)
            y_scrollbar = Scrollbar(self, orient=VERTICAL)
            self._can_container = Canvas(
                self,
                width=self.width,
                height=self.height,
                xscrollcommand=x_scrollbar.set,
                yscrollcommand=y_scrollbar.set,
            )
            self._frm_container = Frame(
                self._can_container, borderwidth=2, relief="groove"
            )
            self._can_container.grid(column=0, row=0, sticky=(N, S, E, W))
            x_scrollbar.grid(column=0, row=1, sticky=(E, W))
            y_scrollbar.grid(column=1, row=0, sticky=(N, S))
            x_scrollbar.config(command=self._can_container.xview)
            y_scrollbar.config(command=self._can_container.yview)
            # ensure container canvas expands to accommodate child frames
            self._can_container.create_window(
                (0, 0), window=self._frm_container, anchor=NW
            )
            self._can_container.bind(
                "<Configure>",
                lambda e: self._can_container.config(
                    scrollregion=self._can_container.bbox(ALL)
                ),
            )
        else:  # normal resolution
            self._frm_container = Frame(self, borderwidth=2, relief="groove")
            self._frm_container.grid(column=0, row=0, sticky=(N, S, E, W))

        # create status frame
        self._frm_status = Frame(self, borderwidth=2, relief="groove")
        self._lbl_status = Label(self._frm_status, textvariable=self._status, anchor=W)
        self._btn_exit = Button(
            self._frm_status,
            image=self.img_exit,
            width=50,
            fg=ERRCOL,
            command=self.on_exit,
        )
        self._frm_status.grid(column=0, row=2, sticky=(W, E))
        self._lbl_status.grid(column=0, row=0, sticky=(W, E))
        self._btn_exit.grid(column=1, row=0, sticky=E)

        # set column and row weights
        # NB!!! these govern the 'pack' behaviour of the frames on resize
        self.grid_columnconfigure(0, weight=10)
        self.grid_rowconfigure(0, weight=10)
        self._frm_status.grid_columnconfigure(0, weight=10)
        if self.lowres:
            colsp, rowsp = self._can_container.grid_size()
        else:
            colsp, rowsp = self._frm_container.grid_size()
        for i in range(colsp):
            self._frm_status.grid_columnconfigure(i, weight=10)
        for i in range(rowsp):
            self._frm_status.grid_rowconfigure(i, weight=10)

    def _finalise(self):
        """
        Finalise Toplevel window after child frames have been created.
        """

        # self.set_status(f"{self.height}, {self.width}") # testing only

    def set_status(self, message: str, color: str = ""):
        """
        Set status message.

        :param str message: message to be displayed
        :param str color: rgb color of text (blue)
        """

        message = (message[:120] + "..") if len(message) > 120 else message
        if color != "":
            self._lbl_status.config(fg=color)
        self._status.set("  " + message)
        self.update_idletasks()

    def on_exit(self, *args, **kwargs):  # pylint: disable=unused-argument
        """
        Handle Exit button press.
        """

        self.__app.stop_dialog(self._dlgname)
        self.destroy()

    def _on_resize(self, event):  # pylint: disable=unused-argument
        """
        Resize frame.

        :param event event: resize event
        """

        self.width, self.height = self.get_size()

    def get_size(self):
        """
        Get current frame size.

        :return: window size (width, height)
        :rtype: tuple
        """

        self.__master.update_idletasks()  # Make sure we know about any resizing
        return self.winfo_width(), self.winfo_height()

    @property
    def container(self):
        """
        Getter for container frame.

        :return: reference to container frame
        :rtype: tkinter.Frame
        """

        return self._frm_container



"""
spartn_mqtt_frame.py

SPARTN configuration dialog

This is the pop-up dialog containing the various
SPARTN configuration functions.

NB: The initial configuration for the SPARTN MQTT client
(pygnssutils.GNSSMQTTClient) is set in app.update_SPARTN_handler().
Once started, the persisted state for the SPARTN MQTT client is
held in the threaded SPARTN handler itself, NOT in this frame.

Created on 26 Jan 2023

:author: semuadmin (Steve Smith)
:copyright: 2020 semuadmin
:license: BSD 3-Clause
"""

from os import path
from pathlib import Path
from tkinter import (
    DISABLED,
    NORMAL,
    Button,
    Checkbutton,
    E,
    Entry,
    Frame,
    IntVar,
    Label,
    Radiobutton,
    Spinbox,
    StringVar,
    TclError,
    W,
    ttk,
)

from PIL import Image, ImageTk
from pyspartn import TIMEBASE

try:
    from pyspartn import HASCRYPTO
except ImportError:
    HASCRYPTO = 1
from pyubx2 import UBXMessage

from pygpsclient.globals import (
    CONNECTED_SPARTNIP,
    CONNECTED_SPARTNLB,
    DISCONNECTED,
    ERRCOL,
    ICON_BLANK,
    ICON_CONFIRMED,
    ICON_DISCONN,
    ICON_EXIT,
    ICON_LOAD,
    ICON_PENDING,
    ICON_SEND,
    ICON_SERIAL,
    ICON_SOCKET,
    ICON_WARNING,
    MQTTIPMODE,
    MQTTLBANDMODE,
    OKCOL,
    READONLY,
    RXMMSG,
    SPARTN_BASEDATE_CURRENT,
    SPARTN_BASEDATE_DATASTREAM,
    SPARTN_GNSS,
    SPARTN_OUTPORT,
    SPARTN_PPREGIONS,
)
from pygpsclient.helpers import MAXPORT, VALINT, VALLEN, valid_entry
from pygpsclient.strings import DLGSPARTNWARN, LBLSPARTNIP, MQTTCONN


class SPARTNMQTTDialog(Frame):
    """,
    SPARTNConfigDialog class.
    """

    def __init__(
        self, app, container, *args, **kwargs
    ):  # pylint: disable=unused-argument
        """
        Constructor.

        :param Frame app: reference to main tkinter application
        :param Frame container: reference to container frame
        :param args: optional args to pass to parent class (not currently used)
        :param kwargs: optional kwargs to pass to parent class (not currently used)
        """

        self.__app = app  # Reference to main application class
        self.__master = self.__app.appmaster  # Reference to root class (Tk)
        self.__container = container  # container frame

        Frame.__init__(self, self.__container.container, *args, **kwargs)

        self._img_blank = ImageTk.PhotoImage(Image.open(ICON_BLANK))
        self._img_pending = ImageTk.PhotoImage(Image.open(ICON_PENDING))
        self._img_confirmed = ImageTk.PhotoImage(Image.open(ICON_CONFIRMED))
        self._img_warn = ImageTk.PhotoImage(Image.open(ICON_WARNING))
        self._img_exit = ImageTk.PhotoImage(Image.open(ICON_EXIT))
        self._img_send = ImageTk.PhotoImage(Image.open(ICON_SEND))
        self._img_serial = ImageTk.PhotoImage(Image.open(ICON_SERIAL))
        self._img_socket = ImageTk.PhotoImage(Image.open(ICON_SOCKET))
        self._img_disconn = ImageTk.PhotoImage(Image.open(ICON_DISCONN))
        self._img_load = ImageTk.PhotoImage(Image.open(ICON_LOAD))
        self._stream_handler = None
        self._status = StringVar()
        self._mqtt_server = StringVar()
        self._mqtt_port = IntVar()
        self._mqtt_port.set(SPARTN_OUTPORT)
        self._mqtt_region = StringVar()
        self._mqtt_source = IntVar()
        self._mqtt_clientid = StringVar()
        self._mqtt_crt = StringVar()
        self._mqtt_pem = StringVar()
        self._mqtt_iptopic = IntVar()
        self._mqtt_mgatopic = IntVar()
        self._mqtt_keytopic = IntVar()
        self._mqtt_freqtopic = IntVar()
        self._spartndecode = IntVar()
        self._settings = {}
        # spartn_inqueue is read by app and passed to GNSS receiver
        self._output = self.__app.spartn_inqueue
        self._connected = DISCONNECTED

        self._body()
        self._do_layout()
        self._reset()
        self._attach_events()

    def _body(self):
        """
        Set up frame and widgets.
        """
        # pylint: disable=unnecessary-lambda

        self._lbl_corrip = Label(self, text=LBLSPARTNIP)
        self._lbl_mqttserver = Label(self, text="Server")
        self._ent_mqttserver = Entry(
            self,
            textvariable=self._mqtt_server,
            state=NORMAL,
            relief="sunken",
            width=25,
        )
        self._lbl_mqttport = Label(self, text="Port")
        self._ent_mqttport = Entry(
            self,
            textvariable=self._mqtt_port,
            state=NORMAL,
            relief="sunken",
            width=8,
        )
        self._lbl_mqttregion = Label(self, text="Region")
        self._spn_mqttregion = Spinbox(
            self,
            values=SPARTN_PPREGIONS,
            textvariable=self._mqtt_region,
            state=READONLY,
            width=4,
            wrap=True,
        )
        self._rad_ip = Radiobutton(
            self, text="IP", variable=self._mqtt_source, value=MQTTIPMODE
        )
        self._rad_lb = Radiobutton(
            self, text="L-Band", variable=self._mqtt_source, value=MQTTLBANDMODE
        )
        self._lbl_mqttclientid = Label(self, text="Client ID")
        self._ent_mqttclientid = Entry(
            self,
            textvariable=self._mqtt_clientid,
            state=NORMAL,
            relief="sunken",
            width=35,
        )
        self._lbl_topics = Label(self, text="Topics")
        self._chk_mqtt_iptopic = Checkbutton(
            self,
            text="IP",
            variable=self._mqtt_iptopic,
        )
        self._chk_mqtt_mgatopic = Checkbutton(
            self,
            text="Assist",
            variable=self._mqtt_mgatopic,
        )
        self._chk_mqtt_keytopic = Checkbutton(
            self,
            text="Key",
            variable=self._mqtt_keytopic,
        )
        self._chk_mqtt_freqtopic = Checkbutton(
            self,
            text="Freq",
            variable=self._mqtt_freqtopic,
        )
        self._btn_opencrt = Button(
            self,
            width=45,
            image=self._img_load,
            command=lambda: self._get_spartncerts("crt"),
        )
        self._lbl_mqttcrt = Label(self, text="CRT File")
        self._ent_mqttcrt = Entry(
            self,
            textvariable=self._mqtt_crt,
            state=NORMAL,
            relief="sunken",
            width=32,
        )
        self._btn_openpem = Button(
            self,
            width=45,
            image=self._img_load,
            command=lambda: self._get_spartncerts("pem"),
        )
        self._lbl_mqttpem = Label(self, text="PEM File")
        self._ent_mqttpem = Entry(
            self,
            textvariable=self._mqtt_pem,
            state=NORMAL,
            relief="sunken",
            width=32,
        )
        self._lbl_spartndecode = Label(self, text="Decode SPARTN in console")
        self._chk_spartndecode = Checkbutton(
            self,
            text="",
            variable=self._spartndecode,
        )
        self._btn_connect = Button(
            self,
            width=45,
            image=self._img_socket,
            command=lambda: self.on_connect(),
        )
        self._btn_disconnect = Button(
            self,
            width=45,
            image=self._img_disconn,
            command=lambda: self.on_disconnect(),
            state=DISABLED,
        )

    def _do_layout(self):
        """
        Position widgets in frame.
        """

        self._lbl_corrip.grid(column=0, row=0, columnspan=6, padx=3, pady=2, sticky=W)
        self._lbl_mqttserver.grid(column=0, row=1, padx=3, pady=2, sticky=W)
        self._ent_mqttserver.grid(
            column=1, row=1, columnspan=5, padx=3, pady=2, sticky=W
        )
        self._lbl_mqttport.grid(column=0, row=2, padx=3, pady=2, sticky=W)
        self._ent_mqttport.grid(column=1, row=2, padx=3, pady=2, sticky=W)
        self._lbl_mqttregion.grid(column=0, row=3, padx=3, pady=2, sticky=W)
        self._spn_mqttregion.grid(column=1, row=3, padx=3, pady=2, sticky=W)
        self._rad_ip.grid(column=2, row=3, padx=3, pady=2, sticky=W)
        self._rad_lb.grid(column=3, row=3, padx=3, pady=2, sticky=W)
        self._lbl_mqttclientid.grid(column=0, row=4, padx=3, pady=2, sticky=W)
        self._ent_mqttclientid.grid(
            column=1, row=4, columnspan=5, padx=3, pady=2, sticky=W
        )
        self._lbl_topics.grid(column=0, row=5, padx=4, pady=2, sticky=W)
        self._chk_mqtt_iptopic.grid(column=1, row=5, padx=3, pady=2, sticky=W)
        self._chk_mqtt_mgatopic.grid(column=2, row=5, padx=3, pady=2, sticky=W)
        self._chk_mqtt_keytopic.grid(column=3, row=5, padx=3, pady=2, sticky=W)
        self._chk_mqtt_freqtopic.grid(column=4, row=5, padx=3, pady=2, sticky=W)
        self._lbl_mqttcrt.grid(column=0, row=6, padx=3, pady=2, sticky=W)
        self._ent_mqttcrt.grid(
            column=1, row=6, padx=3, columnspan=3, pady=2, sticky=(W, E)
        )
        self._btn_opencrt.grid(column=4, row=6, padx=3, pady=2, sticky=E)
        self._lbl_mqttpem.grid(column=0, row=7, padx=3, pady=2, sticky=W)
        self._ent_mqttpem.grid(
            column=1, row=7, columnspan=3, padx=3, pady=2, sticky=(W, E)
        )
        self._btn_openpem.grid(column=4, row=7, padx=3, pady=2, sticky=E)
        self._lbl_spartndecode.grid(column=0, row=8, columnspan=2, sticky=W)
        self._chk_spartndecode.grid(column=2, row=8, sticky=W)
        ttk.Separator(self).grid(
            column=0, row=9, columnspan=6, padx=2, pady=3, sticky=(W, E)
        )
        self._btn_connect.grid(column=0, row=10, padx=3, pady=2, sticky=W)
        self._btn_disconnect.grid(column=2, row=10, padx=3, pady=2, sticky=W)

    def _attach_events(self):
        """
        Set up event listeners.
        """

        self._mqtt_source.trace_add("write", self._on_source)
        for setting in (
            self._mqtt_server,
            self._mqtt_port,
            self._mqtt_region,
            self._mqtt_clientid,
            self._mqtt_crt,
            self._mqtt_pem,
            self._mqtt_iptopic,
            self._mqtt_mgatopic,
            self._mqtt_keytopic,
            self._mqtt_freqtopic,
            self._spartndecode,
        ):
            setting.trace_add("write", self._on_update_config)

    def _reset(self):
        """
        Reset configuration widgets.
        """

        self._get_settings()
        self._reset_keypaths(self._mqtt_clientid.get())
        self.__container.set_status(
            "",
        )
        if self.__app.rtk_conn_status == CONNECTED_SPARTNIP:
            self.set_controls(CONNECTED_SPARTNIP)
        else:
            self.set_controls(DISCONNECTED)

        if not HASCRYPTO:
            self._lbl_spartndecode.config(state=DISABLED)
            self._chk_spartndecode.config(state=DISABLED)

    def _reset_keypaths(self, clientid):
        """
        Reset key and cert file paths.

        :param str clientid: Client ID
        """

        spartn_crt = path.join(Path.home(), f"device-{clientid}-pp-cert.crt")
        spartn_pem = path.join(Path.home(), f"device-{clientid}-pp-key.pem")
        self._mqtt_crt.set(spartn_crt)
        self._mqtt_pem.set(spartn_pem)

    def _on_source(self, var, index, mode):  # pylint: disable=unused-argument
        """
        Action when source is changed.
        """

        self.update()
        cfg = self.__app.configuration
        if self._mqtt_source.get() == MQTTLBANDMODE:
            self._chk_mqtt_iptopic.config(state=DISABLED)
            self._chk_mqtt_freqtopic.config(state=NORMAL)
            self._mqtt_iptopic.set(0)
            self._mqtt_freqtopic.set(1)
        else:  # MQTTIPMODE
            self._chk_mqtt_iptopic.config(state=NORMAL)
            self._chk_mqtt_freqtopic.config(state=DISABLED)
            self._mqtt_iptopic.set(1)
            self._mqtt_freqtopic.set(0)
        cfg.set("mqttclientmode_n", self._mqtt_source.get())

    def _on_update_config(self, var, index, mode):  # pylint: disable=unused-argument
        """
        Update in-memory configuration if setting is changed.
        """

        self.update()
        cfg = self.__app.configuration
        try:
            cfg.set("mqttclientserver_s", self._mqtt_server.get())
            cfg.set("mqttclientport_n", int(self._mqtt_port.get()))
            cfg.set("mqttclientregion_s", self._mqtt_region.get())
            cfg.set("mqttclientid_s", self._mqtt_clientid.get())
            cfg.set("mqttclienttlscrt_s", self._mqtt_crt.get())
            cfg.set("mqttclienttlskey_s", self._mqtt_pem.get())
            cfg.set("mqttclienttopicip_b", int(self._mqtt_iptopic.get()))
            cfg.set("mqttclienttopicmga_b", int(self._mqtt_mgatopic.get()))
            cfg.set("mqttclienttopickey_b", int(self._mqtt_keytopic.get()))
            cfg.set("mqttclienttopicfreq_b", int(self._mqtt_freqtopic.get()))
            cfg.set("spartndecode_b", int(self._spartndecode.get()))
        except (ValueError, TclError):
            pass

    def _get_settings(self):
        """
        Get settings from SPARTN handler.
        """

        self._connected = self.__app.spartn_handler.connected
        if self._connected:
            # get settings from running instance
            self._settings = self.__app.spartn_handler.settings
        else:
            # get settings from saved configuration
            cfg = self.__app.configuration
            self._settings = {}
            self._settings["server"] = cfg.get("mqttclientserver_s")
            self._settings["port"] = cfg.get("mqttclientport_n")
            self._settings["clientid"] = cfg.get("mqttclientid_s")
            self._settings["region"] = cfg.get("mqttclientregion_s")
            self._settings["mode"] = cfg.get("mqttclientmode_n")
            self._settings["topic_ip"] = cfg.get("mqttclienttopicip_b")
            self._settings["topic_mga"] = cfg.get("mqttclienttopicmga_b")
            self._settings["topic_key"] = cfg.get("mqttclienttopickey_b")
            self._settings["topic_freq"] = cfg.get("mqttclienttopicfreq_b")
            self._settings["tlscrt"] = cfg.get("mqttclienttlscrt_s")
            self._settings["tlskey"] = cfg.get("mqttclienttlskey_s")
            self._settings["spartndecode"] = cfg.get("spartndecode_b")
            self._settings["spartnkey"] = cfg.get("spartnkey_s")
            if self._settings["spartnkey"] == "" or not HASCRYPTO:
                self._settings["spartndecode"] = 0
            # if basedate is provided in config file, it must be an integer gnssTimetag
            basedate = cfg.get("spartnbasedate_n")
            if basedate == SPARTN_BASEDATE_CURRENT:
                # pyspartn will interpret 'None' as current datetime
                basedate = None
            elif basedate == SPARTN_BASEDATE_DATASTREAM:
                # pyspartn will interpret 'TIMEBASE' as use gnssTimeTag in datastream
                basedate = TIMEBASE
            self._settings["spartnbasedate"] = basedate

        self._mqtt_server.set(self._settings["server"])
        self._mqtt_port.set(self._settings["port"])
        self._mqtt_region.set(self._settings["region"])
        self._mqtt_source.set(self._settings.get("mode"))
        self._mqtt_clientid.set(self._settings["clientid"])
        self._mqtt_iptopic.set(self._settings["topic_ip"])
        self._mqtt_mgatopic.set(self._settings["topic_mga"])
        self._mqtt_keytopic.set(self._settings["topic_key"])
        self._mqtt_freqtopic.set(self._settings["topic_freq"])
        self._mqtt_crt.set(self._settings["tlscrt"])
        self._mqtt_pem.set(self._settings["tlskey"])
        self._spartndecode.set(self._settings["spartndecode"])

    def _set_settings(self):
        """
        Set settings for SPARTN handler.
        """

        self._settings["server"] = self._mqtt_server.get()
        self._settings["port"] = self._mqtt_port.get()
        self._settings["region"] = self._mqtt_region.get()
        self._settings["mode"] = self._mqtt_source.get()
        self._settings["clientid"] = self._mqtt_clientid.get()
        self._settings["topic_ip"] = self._mqtt_iptopic.get()
        self._settings["topic_mga"] = self._mqtt_mgatopic.get()
        self._settings["topic_key"] = self._mqtt_keytopic.get()
        self._settings["topic_freq"] = self._mqtt_freqtopic.get()
        self._settings["tlscrt"] = self._mqtt_crt.get()
        self._settings["tlskey"] = self._mqtt_pem.get()
        self._settings["spartndecode"] = self._spartndecode.get()
        self._settings["output"] = self._output

    def set_controls(self, status: int):
        """
        Enable or disable MQTT Correction widgets depending on
        connection status.

        :param int status: connection status (0 = disconnected, 1 = connected)
        """

        stat = NORMAL if status == CONNECTED_SPARTNIP else DISABLED
        for wdg in (self._btn_disconnect,):
            wdg.config(state=stat)
        stat = DISABLED if status == CONNECTED_SPARTNIP else NORMAL
        for wdg in (
            self._ent_mqttserver,
            self._ent_mqttport,
            self._btn_connect,
            self._ent_mqttclientid,
            self._ent_mqttcrt,
            self._ent_mqttpem,
            self._rad_ip,
            self._rad_lb,
            self._chk_mqtt_iptopic,
            self._chk_mqtt_mgatopic,
            self._chk_mqtt_keytopic,
            self._chk_mqtt_freqtopic,
            self._btn_opencrt,
            self._btn_openpem,
        ):
            wdg.config(state=stat)
        stat = DISABLED if status == CONNECTED_SPARTNIP else READONLY
        for wdg in (self._spn_mqttregion,):
            wdg.config(state=stat)

        if status == DISCONNECTED:
            self._on_source(None, None, "write")

    def _valid_settings(self) -> bool:
        """
        Validate settings.

        :return: valid True/False
        :rtype: bool
        """

        valid = True
        valid = valid & valid_entry(self._ent_mqttserver, VALLEN, 1, 50)
        valid = valid & valid_entry(self._ent_mqttport, VALINT, 1, MAXPORT)
        valid = valid & valid_entry(self._ent_mqttclientid, VALLEN, 1, 50)
        valid = valid & valid_entry(self._ent_mqttcrt, VALLEN, 1, 254)
        valid = valid & valid_entry(self._ent_mqttpem, VALLEN, 1, 254)
        return valid

    def _get_spartncerts(self, ext: str):
        """
        Get SPARTN TLS key (*.pem) and cert (*.crt) files

        :param str ext: file extension ("crt" or "pem")
        """

        spfile = self.__app.file_handler.open_file(
            "spartncert",
            (
                ("spartn files", f"*.{ext}"),
                ("all files", "*.*"),
            ),
        )
        if ext == "crt":
            self._mqtt_crt.set(spfile)
        elif ext == "pem":
            self._mqtt_pem.set(spfile)

    def on_connect(self):
        """
        Connect to MQTT client.
        """

        if self.__app.rtk_conn_status == CONNECTED_SPARTNLB:
            self.__container.set_status(
                DLGSPARTNWARN.format("L-Band", "IP"),
                ERRCOL,
            )
            return

        server = ""

        if self._valid_settings():
            # if subscribing to SPARTNKEY topic, set pending flag
            # for GNSS dialog to update key details when they arrive
            # via RXM-SPARTNKEY message
            if self._settings["topic_key"]:
                self.__container.set_pending(RXMMSG, SPARTN_GNSS)

            self._set_settings()
            # verbosity and logtofile set in App.__init__()
            self.__app.spartn_handler.start(
                server=self._settings["server"],
                port=self._settings["port"],
                clientid=self._settings["clientid"],
                region=self._settings["region"],
                mode=self._settings["mode"],
                topic_ip=self._settings["topic_ip"],
                topic_mga=self._settings["topic_mga"],
                topic_key=self._settings["topic_key"],
                topic_freq=self._settings["topic_freq"],
                tlscrt=self._settings["tlscrt"],
                tlskey=self._settings["tlskey"],
                spartndecode=self._settings["spartndecode"],
                spartnkey=self._settings["spartnkey"],
                spartnbasedate=self._settings["spartnbasedate"],
                output=self._settings["output"],
            )
            self.set_controls(CONNECTED_SPARTNIP)
            self.__container.set_status(
                MQTTCONN.format(server),
                OKCOL,
            )
            self.__app.rtk_conn_status = CONNECTED_SPARTNIP
        else:
            self.__container.set_status("ERROR! Invalid Settings", ERRCOL)

    def on_disconnect(self, msg: str = ""):
        """
        Disconnect from MQTT client.

        :param str msg: optional disconnection message
        """

        msg += "Disconnected"
        if self.__app.rtk_conn_status == CONNECTED_SPARTNIP:
            self.__app.spartn_handler.stop()
            self.__app.rtk_conn_status = DISCONNECTED
            self.__container.set_status(
                msg,
                ERRCOL,
            )
        self.set_controls(DISCONNECTED)

    def update_status(self, msg: UBXMessage):
        """
        Update pending confirmation status.
        :param UBXMessage msg: UBX config message
        """

    @property
    def clientid(self) -> str:
        """
        Getter for Client ID.

        :return: client ID
        :rtype: str
        """

        return self._mqtt_clientid.get()

    @clientid.setter
    def clientid(self, clientid: str):
        """
        Setter for Client ID.

        :param str clientid: Client ID
        """

        self._mqtt_clientid.set(clientid)
        self._reset_keypaths(clientid)



"""
ubx_handler.py

UBX Protocol handler - handles all incoming UBX messages

Parses individual UBX messages (using pyubx2 library)
and adds selected attribute values to the app.gnss_status
data dictionary. This dictionary is then used to periodically
update the various user-selectable widget panels.

Created on 30 Sep 2020

:author: semuadmin (Steve Smith)
:copyright: 2020 semuadmin
:license: BSD 3-Clause
"""

import logging

from pyubx2 import UBXMessage, itow2utc

from pygpsclient.globals import GLONASS_NMEA, UTF8
from pygpsclient.helpers import corrage2int, fix2desc, ned2vector, svid2gnssid
from pygpsclient.strings import DLGTSPARTN, DLGTUBX, NA
from pygpsclient.widget_state import VISIBLE, WDGSPECTRUM, WDGSYSMON


class UBXHandler:
    """
    UBXHandler class
    """

    def __init__(self, app):
        """
        Constructor.

        :param Frame app: reference to main tkinter application
        """

        self.__app = app  # Reference to main application class
        self.__master = self.__app.appmaster  # Reference to root class (Tk)
        self.logger = logging.getLogger(__name__)

        self._cdb = 0
        self._raw_data = None
        self._parsed_data = None
        # Holds array of current satellites in view from NMEA GSV or UBX NAV-SVINFO sentences
        self.gsv_data = {}

    def process_data(self, raw_data: bytes, parsed_data: object):
        """
        Process relevant UBX message types

        :param bytes raw_data: raw data
        :param UBXMessage parsed_data: parsed data
        """

        if raw_data is None:
            return
        # self.logger.debug(f"data received {parsed_data.identity}")
        if parsed_data.identity[0:3] in ("ACK", "CFG"):
            self._process_ACK(parsed_data)
        elif parsed_data.identity == "ESF-ALG":
            self._process_ESF_ALG(parsed_data)
        elif parsed_data.identity == "HNR-ATT":
            self._process_HNR_ATT(parsed_data)
        elif parsed_data.identity == "HNR-PVT":
            self._process_HNR_PVT(parsed_data)
        elif parsed_data.identity == "NAV-ATT":
            self._process_NAV_ATT(parsed_data)
        elif parsed_data.identity in ("NAV-DOP", "NAV2-DOP"):
            self._process_NAV_DOP(parsed_data)
        elif parsed_data.identity in ("NAV-POSLLH", "NAV-HPPOSLLH"):
            self._process_NAV_POSLLH(parsed_data)
        elif parsed_data.identity in ("NAV-PVT", "NAV2-PVT"):
            self._process_NAV_PVT(parsed_data)
        elif parsed_data.identity == "NAV-PVAT":
            self._process_NAV_PVAT(parsed_data)
        elif parsed_data.identity == "NAV-RELPOSNED":
            self._process_NAV_RELPOSNED(parsed_data)
        elif parsed_data.identity in ("NAV-SAT", "NAV2-SAT"):
            self._process_NAV_SAT(parsed_data)
        elif parsed_data.identity in ("NAV-STATUS", "NAV2-STATUS"):
            self._process_NAV_STATUS(parsed_data)
        elif parsed_data.identity == "NAV-SVIN":
            self._process_NAV_SVIN(parsed_data)
        elif parsed_data.identity == "NAV-SVINFO":
            self._process_NAV_SVINFO(parsed_data)
        elif parsed_data.identity == "NAV-SOL":
            self._process_NAV_SOL(parsed_data)
        elif parsed_data.identity == "NAV-VELNED":
            self._process_NAV_VELNED(parsed_data)
        elif parsed_data.identity == "MON-COMMS":
            self._process_MON_COMMS(parsed_data)
        elif parsed_data.identity == "MON-SPAN":
            self._process_MON_SPAN(parsed_data)
        elif parsed_data.identity == "MON-SYS":
            self._process_MON_SYS(parsed_data)
        elif parsed_data.identity == "MON-VER":
            self._process_MONVER(parsed_data)
        elif parsed_data.identity == "RXM-RTCM":
            self._process_RXM_RTCM(parsed_data)
        elif parsed_data.identity == "RXM-PMP":
            self._process_RXM_PMP(parsed_data)
        elif parsed_data.identity == "RXM-SPARTN-KEY":
            self._process_RXM_SPARTN_KEY(parsed_data)

    def _process_ACK(self, msg: UBXMessage):
        """
        Process ACK-ACK & ACK-NAK sentences and CFG poll responses.

        :param UBXMessage msg: UBX config message
        """

        if self.__app.dialog(DLGTUBX) is not None:
            self.__app.dialog(DLGTUBX).update_pending(msg)

        # if SPARTN config dialog is open, send CFG & ACKs there
        if self.__app.dialog(DLGTSPARTN) is not None:
            self.__app.dialog(DLGTSPARTN).update_pending(msg)

        # if Spectrumview or Sysmon widgets are active, send ACKSs there
        if msg.identity in ("ACK-ACK", "ACK-NAK"):
            wdgs = self.__app.widget_state.state
            for wdg in (WDGSYSMON, WDGSPECTRUM):
                if wdgs[wdg][VISIBLE]:
                    if msg.clsID == 6 and msg.msgID == 1:  # CFG-MSG
                        getattr(self.__app, wdgs[wdg]["frm"]).update_pending(msg)

    def _process_MONVER(self, msg: UBXMessage):
        """
        Process MON-VER sentences.

        :param UBXMessage msg: UBX MON-VER config message
        """

        exts = []
        fw_version = NA
        rom_version = NA
        gnss_supported = ""
        model = ""
        sw_version = getattr(msg, "swVersion", b"N/A")
        sw_version = sw_version.replace(b"\x00", b"").decode(UTF8)
        sw_version = sw_version.replace("ROM CORE", "ROM")
        sw_version = sw_version.replace("EXT CORE", "Flash")
        hw_version = getattr(msg, "hwVersion", b"N/A")
        hw_version = hw_version.replace(b"\x00", b"").decode(UTF8)

        for i in range(9):
            ext = getattr(msg, f"extension_{i+1:02d}", b"")
            ext = ext.replace(b"\x00", b"").decode(UTF8)
            exts.append(ext)
            if "FWVER=" in exts[i]:
                fw_version = exts[i].replace("FWVER=", "")
            if "PROTVER=" in exts[i]:
                rom_version = exts[i].replace("PROTVER=", "")
            if "PROTVER " in exts[i]:
                rom_version = exts[i].replace("PROTVER ", "")
            if "MOD=" in exts[i]:
                model = exts[i].replace("MOD=", "")
                hw_version = f"{model} {hw_version}"
            for gnss in (
                "GPS",
                "GLO",
                "GAL",
                "BDS",
                "SBAS",
                "IMES",
                "QZSS",
                "NAVIC",
            ):
                if gnss in exts[i]:
                    gnss_supported = gnss_supported + gnss + " "

        verdata = {}
        verdata["swversion"] = sw_version
        verdata["hwversion"] = hw_version
        verdata["fwversion"] = fw_version
        verdata["romversion"] = rom_version
        verdata["gnss"] = gnss_supported
        self.__app.gnss_status.version_data = verdata

        if self.__app.dialog(DLGTUBX) is not None:
            self.__app.dialog(DLGTUBX).update_pending(msg)

    def _process_MON_SYS(self, data: UBXMessage):
        """
        Process MON-SYS sentences - System Monitor Information.

        :param UBXMessage data: MON-SYS parsed message
        """

        sysdata = {}
        sysdata["bootType"] = data.bootType
        sysdata["cpuLoad"] = data.cpuLoad
        sysdata["cpuLoadMax"] = data.cpuLoadMax
        sysdata["memUsage"] = data.memUsage
        sysdata["memUsageMax"] = data.memUsageMax
        sysdata["ioUsage"] = data.ioUsage
        sysdata["ioUsageMax"] = data.ioUsageMax
        sysdata["runTime"] = data.runTime
        sysdata["noticeCount"] = data.noticeCount
        sysdata["warnCount"] = data.warnCount
        sysdata["errorCount"] = data.errorCount
        sysdata["tempValue"] = data.tempValue
        self.__app.gnss_status.sysmon_data = sysdata

    def _process_MON_COMMS(self, data: UBXMessage):
        """
        Process MON-COMMS sentences - Comms Port Information.

        :param UBXMessage data: MON-COMMS parsed message
        """

        commsdata = {}
        for i in range(1, data.nPorts + 1):
            idx = f"_{i:02}"
            pid = getattr(data, "portId" + idx)
            tx = getattr(data, "txUsage" + idx)
            txmax = getattr(data, "txPeakUsage" + idx)
            txbytes = getattr(data, "txBytes" + idx)
            txpending = getattr(data, "txPending" + idx)
            rx = getattr(data, "rxUsage" + idx)
            rxmax = getattr(data, "rxPeakUsage" + idx)
            rxbytes = getattr(data, "rxBytes" + idx)
            rxpending = getattr(data, "rxPending" + idx)
            commsdata[pid] = (
                tx,
                txmax,
                txbytes,
                txpending,
                rx,
                rxmax,
                rxbytes,
                rxpending,
            )
        self.__app.gnss_status.comms_data = commsdata

    def _process_NAV_POSLLH(self, data: UBXMessage):
        """
        Process NAV-(HP)POSLLH sentence - Latitude, Longitude, Height.

        :param UBXMessage data: NAV-(HP)POSLLH parsed message
        """

        self.__app.gnss_status.utc = itow2utc(data.iTOW)  # datetime.time
        self.__app.gnss_status.lat = data.lat
        self.__app.gnss_status.lon = data.lon
        self.__app.gnss_status.alt = data.hMSL / 1000  # meters
        self.__app.gnss_status.hacc = data.hAcc / 1000  # meters
        self.__app.gnss_status.vacc = data.vAcc / 1000  # meters
        self.__app.gnss_status.hae = data.height / 1000  # meters

    def _process_NAV_PVT(self, data: UBXMessage):
        """
        Process NAV-PVT sentence -  Navigation position velocity time solution.

        :param UBXMessage data: NAV-PVT parsed message
        """

        self.__app.gnss_status.utc = itow2utc(data.iTOW)  # datetime.time
        self.__app.gnss_status.lat = data.lat
        self.__app.gnss_status.lon = data.lon
        self.__app.gnss_status.alt = data.hMSL / 1000  # meters
        self.__app.gnss_status.hacc = data.hAcc / 1000  # meters
        self.__app.gnss_status.vacc = data.vAcc / 1000  # meters
        self.__app.gnss_status.pdop = data.pDOP
        self.__app.gnss_status.sip = data.numSV
        self.__app.gnss_status.speed = data.gSpeed / 1000  # m/s
        self.__app.gnss_status.track = data.headMot
        self.__app.gnss_status.hae = data.height / 1000  # meters
        self.__app.gnss_status.fix = fix2desc("NAV-PVT", data.fixType)
        if data.carrSoln > 0:
            self.__app.gnss_status.fix = fix2desc("NAV-PVT", data.carrSoln + 5)

        if hasattr(data, "difSoln"):  # for pyubx2 <= 1.2.42
            self.__app.gnss_status.diff_corr = data.difSoln
        elif hasattr(data, "diffSoln"):
            self.__app.gnss_status.diff_corr = data.diffSoln
        if data.lastCorrectionAge != 0:
            self.__app.gnss_status.diff_age = corrage2int(data.lastCorrectionAge)

    def _process_NAV_PVAT(self, data: UBXMessage):
        """
        Process NAV-PVAT sentence -  Navigation position velocity attitude time solution.

        :param UBXMessage data: NAV-PVT parsed message
        """

        self.__app.gnss_status.utc = itow2utc(data.iTOW)  # datetime.time
        self.__app.gnss_status.lat = data.lat
        self.__app.gnss_status.lon = data.lon
        self.__app.gnss_status.alt = data.hMSL / 1000  # meters
        self.__app.gnss_status.speed = data.gSpeed / 1000  # m/s
        self.__app.gnss_status.sip = data.numSV
        self.__app.gnss_status.hae = data.height / 1000  # meters
        ims = self.__app.gnss_status.imu_data
        ims["source"] = data.identity
        ims["roll"] = data.vehRoll
        ims["pitch"] = data.vehPitch
        ims["yaw"] = data.vehHeading
        ims["status"] = (
            (data.vehRollValid << 3) + (data.vehPitchValid << 2) + data.vehHeadingValid
        )

    def _process_NAV_VELNED(self, data: UBXMessage):
        """
        Process NAV-VELNED sentence - Velocity Solution in North East Down format.

        :param UBXMessage data: NAV-VELNED parsed message
        """

        self.__app.gnss_status.track = data.heading
        self.__app.gnss_status.speed = data.gSpeed / 100  # m/s

    def _process_NAV_SAT(self, data: UBXMessage):
        """
        Process NAV-SAT sentences - Space Vehicle Information.

        NB: For consistency with NMEA GSV and UBX NAV-SVINFO message types,
        this uses the NMEA SVID numbering range for GLONASS satellites
        (65 - 96) rather than the Slot ID (1-24) by default.
        To change this, set the GLONASS_NMEA flag in globals.py to False.

        :param UBXMessage data: NAV-SAT parsed message
        """

        show_unused = self.__app.configuration.get("unusedsat_b")
        self.gsv_data = {}
        num_siv = int(data.numSvs)

        for i in range(num_siv):
            idx = f"_{i+1:02d}"
            gnssId = getattr(data, "gnssId" + idx)
            svid = getattr(data, "svId" + idx)
            # use NMEA GLONASS numbering (65-96) rather than slotID (1-24)
            if gnssId == 6 and svid < 25 and svid != 255 and GLONASS_NMEA:
                svid += 64
            elev = getattr(data, "elev" + idx)
            azim = getattr(data, "azim" + idx)
            cno = getattr(data, "cno" + idx)
            if cno == 0 and not show_unused:  # omit unused sats
                continue
            self.gsv_data[f"{gnssId}-{svid}"] = (gnssId, svid, elev, azim, cno)

        self.__app.gnss_status.siv = len(self.gsv_data)
        self.__app.gnss_status.gsv_data = self.gsv_data

    def _process_NAV_STATUS(self, data: UBXMessage):
        """
        Process NAV-STATUS sentences - Status Information.

        :param UBXMessage data: NAV-STATUS parsed message
        """

        self.__app.gnss_status.diff_corr = data.diffSoln
        # self.__app.gnss_status.diff_age = "<60"
        self.__app.gnss_status.fix = fix2desc("NAV-STATUS", data.gpsFix)
        if data.carrSoln > 0:
            self.__app.gnss_status.fix = fix2desc("NAV-STATUS", data.carrSoln + 5)

    def _process_NAV_SVIN(self, data: UBXMessage):
        """
        Process NAV-SVIN sentences - Survey In Status.

        :param UBXMessage data: NAV-SVIN parsed message
        """

        if self.__app.frm_settings.frm_socketserver is not None:
            self.__app.frm_settings.frm_socketserver.svin_countdown(
                data.dur, data.valid, data.active
            )

    def _process_NAV_SVINFO(self, data: UBXMessage):
        """
        Process NAV-SVINFO sentences - Space Vehicle Information.

        NB: Since UBX Gen8 this message is deprecated in favour of NAV-SAT

        :param UBXMessage data: NAV-SVINFO parsed message
        """

        show_unused = self.__app.configuration.get("unusedsat_b")
        self.gsv_data = {}
        num_siv = int(data.numCh)

        for i in range(num_siv):
            idx = f"_{i+1:02d}"
            svid = getattr(data, "svid" + idx)
            gnssId = svid2gnssid(svid)  # derive gnssId from svid
            elev = getattr(data, "elev" + idx)
            azim = getattr(data, "azim" + idx)
            cno = getattr(data, "cno" + idx)
            if cno == 0 and not show_unused:  # omit unused sats
                continue
            self.gsv_data[f"{gnssId}-{svid}"] = (gnssId, svid, elev, azim, cno)

        self.__app.gnss_status.gsv_data = self.gsv_data

    def _process_NAV_SOL(self, data: UBXMessage):
        """
        Process NAV-SOL sentence - Navigation Solution.

        :param UBXMessage data: NAV-SOL parsed message
        """

        self.__app.gnss_status.pdop = data.pDOP
        self.__app.gnss_status.sip = data.numSV
        self.__app.gnss_status.fix = fix2desc("NAV-SOL", data.gpsFix)

    def _process_NAV_DOP(self, data: UBXMessage):
        """
        Process NAV-DOP sentence - Dilution of Precision.

        :param UBXMessage data: NAV-DOP parsed message
        """

        self.__app.gnss_status.pdop = data.pDOP
        self.__app.gnss_status.hdop = data.hDOP
        self.__app.gnss_status.vdop = data.vDOP

    def _process_NAV_RELPOSNED(self, data: UBXMessage):
        """
        Process NAV-RELPOSNED sentence - Relative position of rover.

        Two version of this message:
        - version 0 has no heading or length attributes, so these
          must be derived from the NED values
        - version 1 has heading and length attributes

        NB: pyubx2 parses relPosHP values as mm, so total relPosN
        in cm = relPosN + (relPosHPN * 1e-1), etc.

        :param UBXMessage data: NAV-RELPOSNED parsed message
        """

        if data.version == 0x00:
            n = data.relPosN
            e = data.relPosE
            d = data.relPosD
            relPosLength, relPosHeading = ned2vector(n, e, d)
            n = data.accN * 1e-2
            e = data.accE * 1e-2
            d = data.accD * 1e-2
            accLength, _ = ned2vector(n, e, d)
            accHeading = accLength * relPosHeading / relPosLength  # ballpark
        else:
            relPosLength, relPosHeading, accLength, accHeading = (
                data.relPosLength,
                data.relPosHeading,
                data.accLength * 1e-2,
                data.accHeading,
            )

        self.__app.gnss_status.rel_pos_heading = relPosHeading
        self.__app.gnss_status.rel_pos_length = relPosLength
        self.__app.gnss_status.acc_heading = accHeading
        self.__app.gnss_status.acc_length = accLength
        self.__app.gnss_status.rel_pos_flags = [
            data.gnssFixOK,
            data.diffSoln,
            data.relPosValid,
            data.carrSoln,
            data.isMoving,
            data.refPosMiss,
            data.refObsMiss,
            data.relPosHeadingValid,
            data.relPosNormalized,
        ]

    def _process_HNR_PVT(self, data: UBXMessage):
        """
        Process HNR-PVT sentence -  High Rate Navigation position velocity time solution.

        :param UBXMessage data: HNR-PVT parsed message
        """

        self.__app.gnss_status.utc = itow2utc(data.iTOW)  # datetime.time
        self.__app.gnss_status.lat = data.lat
        self.__app.gnss_status.lon = data.lon
        self.__app.gnss_status.alt = data.hMSL / 1000  # meters
        self.__app.gnss_status.hacc = data.hAcc / 1000  # meters
        self.__app.gnss_status.vacc = data.vAcc / 1000  # meters
        self.__app.gnss_status.speed = data.gSpeed / 1000  # m/s
        self.__app.gnss_status.track = data.headMot
        self.__app.gnss_status.fix = fix2desc("HNR-PVT", data.gpsFix)
        self.__app.gnss_status.diff_corr = data.DiffSoln
        if data.DiffSoln > 0:
            self.__app.gnss_status.fix = fix2desc("HNR-PVT", 6)

    def _process_RXM_RTCM(self, data: UBXMessage):
        """
        Process RXM-RTCM sentences - Status Information.

        :param UBXMessage data: RXM-RTCM parsed message
        """

        self.__app.gnss_status.diff_corr = data.msgUsed >= 1
        self.__app.gnss_status.diff_station = data.refStation

    def _process_MON_SPAN(self, data: UBXMessage):
        """
        Process MON-SPAN sentences - Spectrum Information.

        :param UBXMessage data: MON-SPAN parsed message
        """

        numrf = data.numRfBlocks
        rfbs = []
        for i in range(1, numrf + 1):
            idx = f"_{i:02}"
            spec = getattr(data, "spectrum" + idx)
            spn = getattr(data, "span" + idx)
            res = getattr(data, "res" + idx)
            ctr = getattr(data, "center" + idx)
            pga = getattr(data, "pga" + idx)
            rfbs.append((spec, spn, res, ctr, pga))

        self.__app.gnss_status.spectrum_data = rfbs

    def _process_RXM_SPARTN_KEY(self, data: UBXMessage):
        """
        Process RXM-SPARTN_KEY sentences - poll response.

        :param UBXMessage data: RXM-SPARTN_KEY poll response message
        """

        if self.__app.dialog(DLGTSPARTN) is not None:
            self.__app.dialog(DLGTSPARTN).update_pending(data)

    def _process_RXM_PMP(self, data: UBXMessage):
        """
        Process RXM-PMP sentence - SPARTN L-Band data.

        :param UBXMessage data: RXM-PMP message
        """

        if self.__app.dialog(DLGTSPARTN) is not None:
            self.__app.dialog(DLGTSPARTN).update_pending(data)

    def _process_ESF_ALG(self, data: UBXMessage):
        """
        Process ESF-ALG sentence - External Sensor Fusion IMU Alignment.

        :param UBXMessage data: ESF-ALG message
        """

        ims = self.__app.gnss_status.imu_data
        ims["source"] = data.identity
        ims["roll"] = data.roll
        ims["pitch"] = data.pitch
        ims["yaw"] = data.yaw
        ims["status"] = data.status

    def _process_NAV_ATT(self, data: UBXMessage):
        """
        Process NAV_ATT sentence - Navigation Attitude.

        :param UBXMessage data: NAV_ATT message
        """

        ims = self.__app.gnss_status.imu_data
        ims["source"] = data.identity
        ims["roll"] = data.roll
        ims["pitch"] = data.pitch
        ims["yaw"] = data.heading
        ims["status"] = ""

    def _process_HNR_ATT(self, data: UBXMessage):
        """
        Process HNR_ATT sentence - High Rate Navigation Attitude..

        :param UBXMessage data: HNR_ATT message
        """

        ims = self.__app.gnss_status.imu_data
        ims["source"] = data.identity
        ims["roll"] = data.roll
        ims["pitch"] = data.pitch
        ims["yaw"] = data.heading
        ims["status"] = ""



"""
ubx_config_dialog.py

UBX configuration container dialog

This is the pop-up dialog containing the various
UBX configuration command frames.

Supply initial settings via `config` keyword argument.

NB: Individual UBX configuration commands do not have uniquely
identifiable synchronous or asynchronous responses (e.g. unique
txn ID). The way we keep tabs on confirmation status is to
maintain a list of all commands sent and the responses they're
expecting. When we receive a response, we check against the list
of awaited responses of the same type and flag the first one we
find as 'confirmed'.

Created on 19 Sep 2020

:author: semuadmin (Steve Smith)
:copyright: 2020 semuadmin
:license: BSD 3-Clause
"""

from tkinter import E, N, S, W

from pyubx2 import UBXMessage

from pygpsclient.dynamic_config_frame import Dynamic_Config_Frame
from pygpsclient.globals import (
    CONNECTED,
    CONNECTED_SIMULATOR,
    CONNECTED_SOCKET,
    ENABLE_CFG_LEGACY,
    ERRCOL,
    UBX_CFGMSG,
    UBX_CFGOTHER,
    UBX_CFGPRT,
    UBX_CFGRATE,
    UBX_CFGVAL,
    UBX_MONHW,
    UBX_MONRF,
    UBX_MONVER,
    UBX_PRESET,
)
from pygpsclient.hardware_info_frame import Hardware_Info_Frame
from pygpsclient.strings import DLGTUBX
from pygpsclient.toplevel_dialog import ToplevelDialog
from pygpsclient.ubx_cfgval_frame import UBX_CFGVAL_Frame
from pygpsclient.ubx_msgrate_frame import UBX_MSGRATE_Frame
from pygpsclient.ubx_port_frame import UBX_PORT_Frame
from pygpsclient.ubx_preset_frame import UBX_PRESET_Frame
from pygpsclient.ubx_recorder_frame import UBX_Recorder_Frame
from pygpsclient.ubx_solrate_frame import UBX_RATE_Frame

MINDIM = (570, 1076)


class UBXConfigDialog(ToplevelDialog):
    """,
    UBXConfigDialog class.
    """

    def __init__(self, app, *args, **kwargs):  # pylint: disable=unused-argument
        """
        Constructor.

        :param Frame app: reference to main tkinter application
        :param args: optional args to pass to parent class (not currently used)
        :param kwargs: optional kwargs to pass to parent class (not currently used)
        """

        self.__app = app  # Reference to main application class

        super().__init__(app, DLGTUBX, MINDIM)

        self._cfg_msg_command = None
        self._pending_confs = {}
        self._recordmode = False

        self._body()
        self._do_layout()
        self._reset()
        self._attach_events()
        self._finalise()

    def _body(self):
        """
        Set up frame and widgets.
        """

        # add configuration widgets
        self._frm_device_info = Hardware_Info_Frame(
            self.__app, self, borderwidth=2, relief="groove", protocol="UBX"
        )
        self._frm_recorder = UBX_Recorder_Frame(
            self.__app, self, borderwidth=2, relief="groove"
        )
        self._frm_config_port = UBX_PORT_Frame(
            self.__app, self, borderwidth=2, relief="groove"
        )
        self._frm_config_rate = UBX_RATE_Frame(
            self.__app, self, borderwidth=2, relief="groove"
        )
        self._frm_config_msg = UBX_MSGRATE_Frame(
            self.__app, self, borderwidth=2, relief="groove"
        )
        self._frm_config_dynamic = Dynamic_Config_Frame(
            self.__app, self, borderwidth=2, relief="groove", protocol="UBX"
        )
        self._frm_configdb = UBX_CFGVAL_Frame(
            self.__app, self, borderwidth=2, relief="groove"
        )
        self._frm_preset = UBX_PRESET_Frame(
            self.__app,
            self,
            borderwidth=2,
            relief="groove",
        )

    def _do_layout(self):
        """
        Position widgets in frame.
        """

        # top of grid
        col = 0
        row = 0
        for frm in (
            self._frm_device_info,
            self._frm_recorder,
            self._frm_config_port,
            self._frm_config_rate,
            self._frm_config_msg,
        ):
            colsp, rowsp = frm.grid_size()
            frm.grid(
                column=col,
                row=row,
                columnspan=colsp,
                rowspan=rowsp,
                sticky=(N, S, W, E),
            )
            row += rowsp
        # middle column of grid
        row = 0
        col += colsp
        for frm in (self._frm_configdb, self._frm_preset):
            colsp, rowsp = frm.grid_size()
            frm.grid(
                column=col,
                row=row,
                columnspan=colsp,
                rowspan=rowsp,
                sticky=(N, S, W, E),
            )
            row += rowsp
        # right column of grid
        if ENABLE_CFG_LEGACY:
            row = 0
            col += colsp
            for frm in (self._frm_config_dynamic,):
                (colsp, rowsp) = frm.grid_size()
                frm.grid(
                    column=col,
                    row=row,
                    columnspan=colsp,
                    rowspan=rowsp,
                    sticky=(N, S, W, E),
                )
                row += rowsp

    def _reset(self):
        """
        Reset configuration widgets.
        """

        self._frm_config_rate.reset()
        self._frm_config_port.reset()
        self._frm_config_dynamic.reset()
        self._frm_device_info.reset()
        if self.__app.conn_status not in (
            CONNECTED,
            CONNECTED_SOCKET,
            CONNECTED_SIMULATOR,
        ):
            self.set_status("Device not connected", ERRCOL)

    def _attach_events(self):
        """
        Bind events to window.
        """

        # self.bind("<Configure>", self._on_resize)

    def set_pending(self, msgid: int, ubxfrm: int):
        """
        Set pending confirmation flag for UBX configuration frame to
        signify that it's waiting for a confirmation message.

        :param int msgid: UBX message identity
        :param int ubxfrm: integer representing UBX configuration frame (0-6)
        """

        self._pending_confs[msgid] = ubxfrm

    def update_pending(self, msg: UBXMessage):
        """
        Receives polled confirmation message from the ubx_handler and
        updates whichever UBX config frame is waiting for this confirmation.

        :param UBXMessage msg: UBX config message
        """

        ubxfrm = self._pending_confs.get(msg.identity, None)

        if ubxfrm is not None:
            if ubxfrm in (UBX_MONVER, UBX_MONHW, UBX_MONRF):
                self._frm_device_info.update_status(msg)
            elif ubxfrm == UBX_CFGPRT:
                self._frm_config_port.update_status(msg)
            elif ubxfrm == UBX_CFGRATE:
                self._frm_config_rate.update_status(msg)
            elif ubxfrm == UBX_CFGMSG:
                self._frm_config_msg.update_status(msg)
            elif ubxfrm == UBX_CFGVAL:
                self._frm_configdb.update_status(msg)
            elif ubxfrm == UBX_PRESET:
                self._frm_preset.update_status(msg)
            elif ubxfrm == UBX_CFGOTHER:
                self._frm_config_dynamic.update_status(msg)

            # reset all confirmation flags for this frame
            for msgid in (msg.identity, "ACK-ACK", "ACK-NAK"):
                if self._pending_confs.get(msgid, None) == ubxfrm:
                    self._pending_confs.pop(msgid)

    @property
    def recordmode(self) -> bool:
        """
        Getter for recording status.

        :return: recording yes/no
        :rtype: bool
        """

        return self._recordmode

    @recordmode.setter
    def recordmode(self, recordmode: bool):
        """
        Setter for record mode.

        :param bool recordmode: recording yes/no
        """

        self._recordmode = recordmode

    def send_command(self, msg: UBXMessage):
        """
        Send command to receiver.
        """

        self.__app.gnss_outqueue.put(msg.serialize())
        self.record_command(msg)

    def record_command(self, msg: UBXMessage):
        """
        Record command to memory if in 'record' mode.
        """

        if self.recordmode:
            self._frm_recorder.update_record(msg)



"""
mapquest.py

MapQuest API Constants and Methods.

MapQuest polygon compression and decompression routines
adapted from the original javascript examples:

https://developer.mapquest.com/documentation/api/static-map/
https://developer.mapquest.com/documentation/common/encode-decode/

Created on 04 May 2023

:author: semuadmin (Steve Smith)
:copyright: 2020 semuadmin
:license: BSD 3-Clause

"""

from pygpsclient.globals import Area

# MapQuest API URLS:
MAPQURL = (
    "https://www.mapquestapi.com/staticmap/v5/map?"
    + "key={key}&type={type}&size={width},{height}"
)
BBOX = "&boundingBox={lat2},{lon1},{lat1},{lon2}"  # top left, bottom right
HACC = "&shape=radius:{radius}|weight:1|fill:ccffff50|border:88888850|{lat},{lon}"
LOC = "&locations={lat},{lon}|{marker}"
LOCS = "&locations={lat1},{lon1}||{lat2},{lon2}&defaultMarker=marker-num"
LOCICON = "marker-sm-616161-ff4444"
MARGIN = "&margin={margin}"  # bbox margin (default 50)
RETINA = "@2x"  # append to size clause for double the resolution
SCALE = "&scalebar={scale}|bottom"  # true or false
TRK = "&shape=weight:2|border:{color}|{track}"
ZOOM = "&zoom={zoom}"  # zoom level 1-20

POINTLIMIT = 500  # max number of shape points supported by MapQuest API
MAPQTIMEOUT = 5
# how frequently the mapquest api is called to update the web map (seconds)
MAP_UPDATE_INTERVAL = 60
MIN_UPDATE_INTERVAL = 5
MAX_ZOOM = 20
MIN_ZOOM = 1
TRKCOL = "ff00ff"
# MapQuest static API map types ("dark" and "light" not used here)
MAP = "map"
SAT = "sat"
HYB = "hyb"


def compress_track(track: tuple, precision: int = 6, limit: int = POINTLIMIT) -> str:
    """
    Convert track to compressed Mapquest format.

    :param tuple track: tuple of Points
    :param int precision: no decimal places precision (6)
    :param int limit: max no of points (500)
    :return: compressed track
    :rtype: str
    """

    # if the number of trackpoints exceeds the MapQuest API limit,
    # increase step count until the number is within limits
    points = []
    stp = 1
    rng = len(track)
    while rng / stp > limit:
        stp += 1
    for i, p in enumerate(track):
        if i % stp == 0:
            points.append(p.lat)
            points.append(p.lon)

    # compress polygon for MapQuest API
    return mapq_compress(points, precision)


def format_mapquest_request(
    mqapikey: str,
    maptype: str,
    width: int,
    height: int,
    zoom: int,
    locations: list,
    bbox: Area = None,
    hacc: float = 0,
) -> str:
    """
    Formats URL for web map download.

    :param str mqapikey: MapQuest API key
    :param str maptype: "map" or "sat"
    :param int width: width of canvas
    :param int height: height of canvas
    :param int zoom: zoom factor
    :param list locations: list of Points
    :param Area bbox: bounding box (will override zoom)
    :param float hacc: horizontal accuracy
    :return: formatted MapQuest URL
    :rtype: str
    """
    # pylint: disable=too-many-arguments, too-many-positional-arguments

    url = MAPQURL.format(key=mqapikey, type=maptype, width=width, height=height)
    radius = str(hacc / 1000)  # km
    zoom = min(20, zoom)

    if bbox is None:  # use location and zoom level
        url += ZOOM.format(zoom=zoom)
    else:  # use bounding box (remember upper left, bottom right)
        url += BBOX.format(
            lat2=bbox.lat2, lon1=bbox.lon1, lat1=bbox.lat1, lon2=bbox.lon2
        ) + MARGIN.format(margin=0)

    if isinstance(locations, list):  # at least one location
        if len(locations) > 1:  # multiple locations (track)
            comp = compress_track(locations)
            url += LOCS.format(
                lat1=locations[0].lat,
                lon1=locations[0].lon,
                lat2=locations[-1].lat,
                lon2=locations[-1].lon,
            ) + TRK.format(color=TRKCOL, track=f"cmp6|enc:{comp}")
        else:  # single location
            url += LOC.format(
                lat=locations[0].lat, lon=locations[0].lon, marker=LOCICON
            )

        if hacc > 0:
            url += HACC.format(
                radius=radius, lat=locations[0].lat, lon=locations[0].lon
            )

    # seems to be bug in MapQuest API which causes error
    # if scalebar displayed at maximum zoom
    url += SCALE.format(scale=("true" if zoom < 20 else "false"))

    return url


def mapq_encode(num: int) -> str:
    """
    Encode number representing character.

    :param int num: number to encode
    :return: encoded number as string
    :rtype: str
    """

    num = num << 1
    if num < 0:
        num = ~(num)

    encoded = ""
    while num >= 0x20:
        encoded += chr((0x20 | (num & 0x1F)) + 63)
        num >>= 5

    encoded += chr(num + 63)
    return encoded


def mapq_decompress(encoded: str, precision: int = 6) -> list:
    """
    Decompress polygon for MapQuest API.

    :param str encoded: polygon encoded as string
    :param int precision: no decimal places precision (6)
    :return: polygon as list of point tuples (lat,lon)
    :rtype: list
    """

    precision = 10**-precision
    leng = len(encoded)
    index = 0
    lat = 0
    lng = 0
    array = []
    while index < leng:
        shift = 0
        result = 0
        b = 0xFF
        while b >= 0x20:
            b = ord(encoded[index]) - 63
            index += 1
            result |= (b & 0x1F) << shift
            shift += 5

        dlat = ~(result >> 1) if (result & 1) else (result >> 1)
        lat += dlat
        shift = 0
        result = 0
        b = 0xFF
        while b >= 0x20:
            b = ord(encoded[index]) - 63
            index += 1
            result |= (b & 0x1F) << shift
            shift += 5

        dlng = ~(result >> 1) if (result & 1) else (result >> 1)
        lng += dlng
        array.append(lat * precision)
        array.append(lng * precision)

    return array


def mapq_compress(points: list, precision: int = 6) -> str:
    """
    Compress polygon for MapQuest API.

    :param list points: polygon as list of point tuples (lat, lon)
    :param int precision: no decimal places precision (6)
    :return: polygon encoded as string
    :rtype: string
    """

    oldLat = 0
    oldLng = 0
    leng = len(points)
    index = 0
    encoded = ""
    precision = 10**precision
    while index < leng:
        #  Round to N decimal places
        lat = round(points[index] * precision)
        index += 1
        lng = round(points[index] * precision)
        index += 1

        #  Encode the differences between the points
        encoded += mapq_encode(lat - oldLat)
        encoded += mapq_encode(lng - oldLng)

        oldLat = lat
        oldLng = lng

    return encoded



"""
ubx_solrate_frame.py

UBX Configuration frame for CFG-RATE commands.

Created on 19 Feb 2020

:author: semuadmin (Steve Smith)
:copyright: 2020 semuadmin
:license: BSD 3-Clause
"""

from tkinter import Button, E, Frame, IntVar, Label, Spinbox, StringVar, W

from PIL import Image, ImageTk
from pyubx2 import POLL, SET, UBXMessage

from pygpsclient.globals import (
    CONNECTED,
    ERRCOL,
    ICON_CONFIRMED,
    ICON_PENDING,
    ICON_SEND,
    ICON_WARNING,
    OKCOL,
    READONLY,
    UBX_CFGRATE,
)
from pygpsclient.strings import LBLCFGRATE

TIMEREFS = {
    0: "UTC",
    1: "GPS",
    2: "GLO",
    3: "BEI",
    4: "GAL",
}


class UBX_RATE_Frame(Frame):
    """
    UBX Navigation Solution Rate configuration command panel.
    """

    def __init__(self, app, container, *args, **kwargs):
        """
        Constructor.

        :param Frame app: reference to main tkinter application
        :param Frame container: reference to container frame (config-dialog)
        :param args: optional args to pass to Frame parent class
        :param kwargs: optional kwargs to pass to Frame parent class
        """

        self.__app = app  # Reference to main application class
        self.__master = self.__app.appmaster  # Reference to root class (Tk)
        self.__container = container

        super().__init__(container.container, *args, **kwargs)

        self._img_send = ImageTk.PhotoImage(Image.open(ICON_SEND))
        self._img_pending = ImageTk.PhotoImage(Image.open(ICON_PENDING))
        self._img_confirmed = ImageTk.PhotoImage(Image.open(ICON_CONFIRMED))
        self._img_warn = ImageTk.PhotoImage(Image.open(ICON_WARNING))
        self._measint = IntVar()
        self._navrate = IntVar()
        self._timeref = StringVar()

        self._body()
        self._do_layout()
        self._attach_events()

    def _body(self):
        """
        Set up frame and widgets.
        """

        self._lbl_cfg_rate = Label(self, text=LBLCFGRATE, anchor=W)
        self._lbl_ubx_measint = Label(self, text="Solution Interval (ms)")
        self._spn_ubx_measint = Spinbox(
            self,
            values=(25, 50, 100, 200, 250, 500, 1000, 2000, 5000, 10000),
            width=6,
            state=READONLY,
            wrap=True,
            textvariable=self._measint,
        )
        self._lbl_ubx_navrate = Label(self, text="Measurement Ratio")
        self._spn_ubx_navrate = Spinbox(
            self,
            from_=1,
            to=127,
            width=4,
            state=READONLY,
            wrap=True,
            textvariable=self._navrate,
        )
        self._lbl_ubx_timeref = Label(self, text="Time Reference")
        self._spn_ubx_timeref = Spinbox(
            self,
            values=list(TIMEREFS.values()),
            width=5,
            state=READONLY,
            wrap=True,
            textvariable=self._timeref,
        )
        self._lbl_send_command = Label(self, image=self._img_pending)
        self._btn_send_command = Button(
            self,
            image=self._img_send,
            width=50,
            command=self._on_send_rate,
            font=self.__app.font_md,
        )

    def _do_layout(self):
        """
        Layout widgets.
        """

        self._lbl_cfg_rate.grid(column=0, row=0, columnspan=6, padx=3, sticky=(W, E))
        self._lbl_ubx_measint.grid(
            column=0, row=1, columnspan=2, rowspan=1, padx=3, pady=3, sticky=W
        )
        self._spn_ubx_measint.grid(column=2, row=1, columnspan=1, rowspan=1, sticky=W)
        self._lbl_ubx_navrate.grid(
            column=0, row=2, columnspan=2, rowspan=1, padx=3, pady=3, sticky=W
        )
        self._spn_ubx_navrate.grid(column=2, row=2, columnspan=2, rowspan=1, sticky=W)
        self._lbl_ubx_timeref.grid(
            column=0, row=3, columnspan=2, rowspan=1, padx=3, pady=3, sticky=W
        )
        self._spn_ubx_timeref.grid(column=2, row=3, columnspan=2, rowspan=1, sticky=W)
        self._btn_send_command.grid(
            column=4, row=1, rowspan=3, ipadx=3, ipady=3, sticky=E
        )
        self._lbl_send_command.grid(
            column=5, row=1, rowspan=3, ipadx=3, ipady=3, sticky=E
        )

        (cols, rows) = self.grid_size()
        for i in range(cols):
            self.grid_columnconfigure(i, weight=1)
        for i in range(rows):
            self.grid_rowconfigure(i, weight=1)
        self.option_add("*Font", self.__app.font_sm)

    def _attach_events(self):
        """
        Bind events to widget.
        """

        # click mouse button to refresh information
        self.bind("<Button>", self._do_poll_rate)

    def reset(self):
        """
        Reset panel to initial settings.
        """

        if self.__app.conn_status == CONNECTED:
            self._do_poll_rate()

    def update_status(self, msg: UBXMessage):
        """
        Update pending confirmation status.

        :param UBXMessage msg: UBX config message
        """

        if msg.identity == "CFG-RATE":
            self._measint.set(msg.measRate)
            self._navrate.set(msg.navRate)
            self._timeref.set(TIMEREFS[msg.timeRef])
            self._lbl_send_command.config(image=self._img_confirmed)
            self.__container.set_status("CFG-RATE GET message received", OKCOL)

        elif msg.identity == "ACK-NAK":
            self.__container.set_status("CFG-RATE POLL message rejected", ERRCOL)
            self._lbl_send_command.config(image=self._img_warn)

    def _on_send_rate(self, *args, **kwargs):  # pylint: disable=unused-argument
        """
        Handle Send rate config button press.
        """

        tref = 0
        for key, val in TIMEREFS.items():
            if val == self._timeref.get():
                tref = key
                break

        msg = UBXMessage(
            "CFG",
            "CFG-RATE",
            SET,
            measRate=self._measint.get(),
            navRate=self._navrate.get(),
            timeRef=tref,
        )
        self.__container.send_command(msg)
        self._lbl_send_command.config(image=self._img_pending)
        self.__container.set_status(
            "CFG-RATE SET message sent",
        )
        self.__container.set_pending(UBX_CFGRATE, ("ACK-ACK", "ACK-NAK"))

        self._do_poll_rate()

    def _do_poll_rate(self, *args, **kwargs):  # pylint: disable=unused-argument
        """
        Poll RATE message configuration.
        """

        msg = UBXMessage("CFG", "CFG-RATE", POLL)
        self.__container.send_command(msg)
        self._lbl_send_command.config(image=self._img_pending)
        self.__container.set_status(
            "CFG-RATE POLL message sent",
        )
        for msgid in ("CFG-RATE", "ACK-NAK"):
            self.__container.set_pending(msgid, UBX_CFGRATE)
